# fmt: off

from .__ import *
from .base import *


__all__ = [
    'IteratorData',
]


################################################################################

type virtual_iterator = iterator | async_iterator | generator | async_iterator

################################################################################

class IteratorData(ResourceData):
    __slots__ = 'is_gen', 'is_async',

    FORBIDDEN_FORM = forbidden_object
    is_gen: bool
    is_async: bool

    @classmethod
    def describe_resource(cls, it_obj: IteratorT) -> 'IteratorData':
        data = IteratorData()
        data.is_gen = isinstance(it_obj, A.Generator)
        data.is_async = isinstance(it_obj, A.AsyncGenerator)
        return data

    def create_resource(self, handle: ResourceHandle) -> virtual_iterator:
        if self.is_gen:
            handle.kind = Kind.Generator
            vobj = async_generator() if self.is_async else generator()
        else:
            handle.kind = Kind.Iterator
            vobj = async_iterator() if self.is_async else iterator()
        handle.keys = []
        handle.open = False
        handle.name = f'<{handle.kind}>'
        obj_set(vobj, VHDL, handle)
        return vobj


################################################################################

class iterator(A.Iterator):
    __slots__ = VHDL,

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> Any:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallSystemMethod(self, next))

    def __str__(self) -> str:
        return '<iterator object>'

    __repr__ = __str__

iterator.__module__ = 'virtual'


################################################################################

class async_iterator(A.AsyncIterator):
    __slots__ = VHDL,

    def __aiter__(self) -> Self:
        return self

    def __anext__(self) -> Any:
        handle = obj_handle(self)
        return handle.hdlr(handle, ResourceCallSystemMethod(self, next).as_coro)

    def __str__(self) -> str:
        return '<async iterator object>'

    __repr__ = __str__

async_iterator.__module__ = 'virtual'


################################################################################

class generator(A.Generator):

    __slots__ = VHDL,

    def __next__(self) -> Any:
        handle = obj_handle(self)
        request = ResourceCallSystemMethod(self, next)
        return handle.hdlr(handle, request)

    def send(self, arg: Any, /) -> Any:
        handle = obj_handle(self)
        request = ResourceCallMethod(self, 'send', (arg,))
        return handle.hdlr(handle, request)

    def throw(self, typ, val=ARG_DEFAULT, tb=ARG_DEFAULT, /) -> Any:
        handle = obj_handle(self)
        request = ResourceCallMethod(self, 'throw', (typ, val, tb))
        return handle.hdlr(handle, request)

    def close(self) -> None:
        handle = obj_handle(self)
        request = ResourceCallMethod(self, 'close')
        return handle.hdlr(handle, request)

    def __str__(self) -> str:
        return '<generator object>'

    __repr__ = __str__

generator.__module__ = 'virtual'


################################################################################

class async_generator(async_iterator, A.AsyncGenerator):

    __slots__ = VHDL,

    def __anext__(self) -> Any:
        handle = obj_handle(self)
        request = ResourceCallSystemMethod(self, next).set_async_mode('coro')
        return handle.hdlr(handle, request)

    def asend(self, arg: Any, /) -> Any:
        handle = obj_handle(self)
        request = ResourceCallMethod(self, 'send', (arg, )).set_async_mode('coro')
        return handle.hdlr(handle, request)

    def athrow(self, typ, val=ARG_DEFAULT, tb=ARG_DEFAULT, /) -> Any:
        handle = obj_handle(self)
        request = ResourceCallMethod(self, 'throw', (typ, val, tb)).set_async_mode('coro')
        return handle.hdlr(handle, request)

    def aclose(self) -> Any:
        handle = obj_handle(self)
        request = ResourceCallMethod(self, 'close').set_async_mode('coro')
        return handle.hdlr(handle, request)

    def __str__(self) -> str:
        return '<async generator object>'

    __repr__ = __str__

generator.__module__ = 'virtual'



# fmt: off

from pathlib import Path
from types import TracebackType

__all__ = [
    'sanitize_exception_for_repl',
    'sanitize_traceback_for_repl',
    'tb_slice',
    'register_repl_path'
]


################################################################################

def sanitize_exception_for_repl(exc: BaseException | None, /) -> BaseException | None:
    if not isinstance(exc, BaseException):
        return exc
    try:
        exc.__traceback__ = sanitize_traceback_for_repl(exc.__traceback__)
    except AttributeError:
        pass
    try:
        exc.__cause__ = sanitize_exception_for_repl(exc.__cause__)
    except AttributeError:
        pass
    try:
        exc.__context__ = sanitize_exception_for_repl(exc.__context__)
    except AttributeError:
        pass
    return exc


################################################################################

MAX_FRAMES = 128

def sanitize_traceback_for_repl(tb: TracebackType | None) -> TracebackType | None:

    if not isinstance(tb, TracebackType):
        return tb

    # find the first non-repl-internal frame
    tb_first = tb
    while tb_first and tb_in_repl(tb_first):
        tb_first = tb_first.tb_next

    # not sure how this could happen
    if tb_first is None:
        return None

    # find the first agentica-internal frame (e.g. warp callback)
    tb_last = tb_first
    i = 0
    while tb_last and not tb_in_agentica(tb_last) and i < MAX_FRAMES:
        tb_last = tb_last.tb_next
        i += 1

    # there was no agentica frame: we just return the tail
    if tb_last is None:
        return tb_first

    # there was a agentica frame: we must create a new chain between tb_first and tb_last
    return tb_slice(tb_first, tb_last)

################################################################################

def tb_slice(tb_first: TracebackType, tb_last: TracebackType | None) -> TracebackType | None:

    try:
        if tb_first is tb_last:
            return None

        if tb_last is None:
            return tb_first

        if tb_first is None:
            return None

        tb_root = tb_copy(tb_first)
        tb_leaf = tb_root
        tb_next = tb_first.tb_next
        while tb_next and tb_next is not tb_last:
            tb_leaf.tb_next = tb_copy(tb_next)
            tb_leaf = tb_leaf.tb_next
            tb_next = tb_next.tb_next

        return tb_root

    except Exception:
        return None

################################################################################

def tb_copy(tb: TracebackType) -> TracebackType:
    return TracebackType(
        tb_next=None,
        tb_frame=tb.tb_frame,
        tb_lasti=tb.tb_lasti,
        tb_lineno=tb.tb_lineno
    )


################################################################################

def tb_in_repl(tb: TracebackType) -> bool:
    filename = tb.tb_frame.f_code.co_filename
    return filename.startswith(REPL_PATHS) or '/sandbox/guest/' in filename

def tb_in_agentica(tb: TracebackType) -> bool:
    filename = tb.tb_frame.f_code.co_filename
    return filename.startswith(AGENTICA_PATHS) or '/agentica_internal/' in filename


################################################################################

# NOTE TO TSLIL: we need to figure out how this changes when packaged

THIS_FILE = Path(__file__)
REPL_PATHS = (THIS_FILE.parent.as_posix(), )         # for agentica_internal.repl
AGENTICA_PATHS = (THIS_FILE.parent.parent.as_posix(), )  # for agentica_internal

def register_repl_path(path: str) -> None:
    global REPL_PATHS, AGENTICA_PATHS
    if path.endswith('.py'):
        path = Path(path).parent.as_posix()
    REPL_PATHS = REPL_PATHS + (path,)
    if not path.startswith(AGENTICA_PATHS):
        AGENTICA_PATHS = AGENTICA_PATHS + (path,)



# fmt: off

from types import ModuleType, MethodType
from .exceptions import ForbiddenError

__all__ = [
    'is_forbidden',
    'blacklist_modules',
    'whitelist_modules',
    'whitelist_objects',
    'forbidden_class',
    'forbidden_object',
    'forbidden_function',
    'forbidden_module',
    'ForbiddenError',
]

################################################################################

BLACKLIST_MODULES = 'agentica_internal.warpc.', 'virtual.', 'agentica'
WHITELIST_MODULES = 'agentica_internal.testing.',
WHITELIST_OBJECTS: set[int] = set()

################################################################################

def blacklist_modules(*mod_names: str) -> None:
    global BLACKLIST_MODULES
    BLACKLIST_MODULES = tuple(set(BLACKLIST_MODULES + mod_names))

def whitelist_modules(*mod_names: str) -> None:
    global WHITELIST_MODULES
    WHITELIST_MODULES = tuple(set(WHITELIST_MODULES + mod_names))

def whitelist_objects(*objs: object) -> None:
    add = WHITELIST_OBJECTS.add
    for obj in objs:
        if type(obj) is MethodType:
            cls, fun = obj.__self__, obj.__func__
            add(id(cls))
            add(id(fun))
        else:
            add(id(obj))

def is_forbidden(obj: object, mod_name: str | None) -> bool:
    return (
        type(mod_name) is str and
        mod_name.startswith(BLACKLIST_MODULES) and
        not mod_name.startswith(WHITELIST_MODULES) and
        id(obj) not in WHITELIST_OBJECTS
    )

################################################################################

fobj = '<forbidden object>'
fcls = '<forbidden class>'
ffun = '<forbidden function>'
fmod = '<forbidden module>'
ffil = '<forbidden file>'


class forbidden_class:
    def __init__(self, *args, **kwargs) -> None:
        raise ForbiddenError('tried to instantiate a forbidden class')

    def __init_subclass__(cls, **kwargs):
        raise ForbiddenError('tried to subclass a forbidden class')

    def __class_getitem__(cls, item):
        raise ForbiddenError('tried to parameterize a forbidden class')

    def __getattr__(self, name: str):
        raise ForbiddenError('tried to get an attribute on a forbidden object')

    def __setattr__(self, name: str, value) -> None:
        raise ForbiddenError('tried to set an attribute on a forbidden object')

    def __delattr__(self, name: str):
        raise ForbiddenError('tried to delete an attribute on a forbidden object')

    def __str__(self) -> str:
        return fobj

    def __repr__(self) -> str:
        return fobj


forbidden_class.__name__ = forbidden_class.__qualname__ = fcls
forbidden_class.__module__ = fmod

################################################################################

forbidden_object = object.__new__(forbidden_class)

################################################################################


def forbidden_function(*args, **kwargs):
    raise ForbiddenError('tried to call a forbidden function')


forbidden_function.__code__ = forbidden_function.__code__.replace(co_name=ffun, co_qualname=ffun)
forbidden_function.__name__ = forbidden_function.__qualname__ = ffun
forbidden_function.__module__ = fmod
forbidden_function.__file__ = ffil

################################################################################


class ForbiddenModule(ModuleType):
    def __getattr__(self, name: str):
        raise ForbiddenError('tried to get an attribute on a forbidden module')

    def __setattr__(self, name: str, value) -> None:
        raise ForbiddenError('tried to set an attribute on a forbidden module')

    def __delattr__(self, name: str):
        raise ForbiddenError('tried to delete an attribute on a forbidden module')

    def __dir__(self):
        return []


forbidden_module = ModuleType(fmod)
forbidden_module.__file__ = ffil



# fmt: off

from .__ import *
from .base import *

__all__ = [
    'FunctionData',
]


################################################################################

WARP_OVERLOAD_MAGIC = hash('warp_overloads') << 32


def register_overload(func: FunctionT, overload_fn: FunctionT):
    """register a virtual warp overload if non-warp overloads not already present"""
    import typing
    from collections import defaultdict

    # {module: {qualname: {firstlineno: func}}}
    overloads: defaultdict[str, defaultdict[str, dict[int, FunctionT]]] | None
    overloads = getattr(typing, '_overload_registry', None) # type: ignore

    if overloads is not None:
        # register overload if not already overloads for this function
        f = getattr(func, '__func__', func)
        o_fn = getattr(overload_fn, '__func__', overload_fn)
        # overloads are distinguished by their line number.
        # to avoid any virtual overloads from clashing with the real overloads,
        # we OR the line number with a magic value.
        lineno = o_fn.__code__.co_firstlineno | WARP_OVERLOAD_MAGIC
        linenos = overloads[f.__module__][f.__qualname__].keys()
        if all(_is_warp_overload(l) for l in linenos):
            overloads[f.__module__][f.__qualname__][lineno] = overload_fn


def retrieve_overloads(func: FunctionT) -> tuple[FunctionT, ...]:
    """retrieve genuine non-warp overloads for a function"""
    import typing
    from collections import defaultdict

    # {module: {qualname: {firstlineno: func}}}
    overloads: defaultdict[str, defaultdict[str, dict[int, FunctionT]]] | None
    overloads = getattr(typing, '_overload_registry', None) # type: ignore

    if overloads is None:
        return ()

    f = getattr(func, '__func__', func)
    if f.__module__ not in overloads:
        return ()
    mod_dict = overloads[f.__module__]
    if f.__qualname__ not in mod_dict:
        return ()

    matches = mod_dict[f.__qualname__].items()
    return tuple(o for l, o in matches if not _is_warp_overload(l))


def _is_warp_overload(lineno: int) -> bool:
    return lineno & WARP_OVERLOAD_MAGIC == WARP_OVERLOAD_MAGIC

################################################################################

class FunctionData(ResourceData):
    __slots__ = (
        'name', 'qname', 'module', 'lineno', 'doc', 'keys', 'annos', 'defaults',
        'args', 'pos_args', 'key_args', 'opt_args', 'pos_star', 'key_star',
        'sig', 'owner', 'overloads', 'is_async', 'async_mode'
    )

    FORBIDDEN_FORM = forbidden_function

    name:      str
    qname:     optstr
    module:    optstr
    lineno:    int
    doc:       optstr
    keys:      strtup
    annos:     AnnotationsT
    defaults:  AttributesT

    args:      strtup
    pos_args:  strtup
    key_args:  strtup
    opt_args:  strtup
    pos_star:  optstr
    key_star:  optstr

    sig:       object | None
    owner:     type | None
    overloads: Tup['FunctionData']

    is_async:   bool
    async_mode: AsyncMode

    # implementation attached later
    @classmethod
    def describe_resource(cls, fun: FunctionT, is_overload: bool = False) -> 'FunctionData': ...

    # implementation attached later
    def create_resource(self, handle: ResourceHandle) -> FunctionT: ...


################################################################################

def describe_real_function(function: FunctionT,
                           is_overload: bool = False,
                           process_defaults: bool = True) -> FunctionData:
    from agentica_internal.cpython.function import func_sig_info
    from agentica_internal.cpython.inspect import (callable_module_and_name,
                                               has_coroutine_mark,
                                               resolve_callable)

    data = FunctionData()

    fun, skip = resolve_callable(function)

    info = func_sig_info(fun)
    pos_args, add_pos = mklist()
    reg_args, add_reg = mklist()
    key_args, add_key = mklist()
    opt_args, add_opt = mklist()
    for arg in info.arg_info[skip:]:
        name = arg.name
        if arg.pos_only:
            add_pos(name)
        elif arg.key_only:
            add_key(name)
        else:
            add_reg(name)
        if arg.optional:
            add_opt(name)

    data.args = tuple(reg_args)
    data.pos_args = tuple(pos_args)
    data.key_args = tuple(key_args)
    data.opt_args = tuple(opt_args)

    data.pos_star = info.pos_star
    data.key_star = info.key_star
    data.is_async = is_async = info.is_async | has_coroutine_mark(fun)
    data.owner    = None
    # for now, this will be set later during FunctionData.encode_fields, where the owning class
    # is known via `.enc_owner()`.

    async_mode = flags.DEFAULT_ASYNC_MODE if is_async else None

    mname, qname, name = callable_module_and_name(fun)
    if mname == '__main__':
        mname = None
    if qname == name:
        qname = None

    data.module, data.qname, data.name = mname, qname, name

    if is_forbidden(function, mname):
        raise E.WarpEncodingForbiddenError(f"<function '{mname}.{qname}'>")

    data.lineno = 0
    if co := getattr(fun, '__code__', None):
        data.lineno = co.co_firstlineno

    fdict, doc, annos = multi_get_raw(fun, DICT, DOC, ANNOS)
    data.keys = tuple(fdict.keys()) if is_rec(fdict) else ()
    data.doc = doc if is_optstr(doc) else None
    data.annos = annos if is_rec(annos) else {}

    defaults_flag: Any = flags.VIRTUAL_FUNCTION_DEFAULTS if process_defaults else 'all'
    data.defaults = get_fun_defaults(fun, data.opt_args, defaults_flag)
    data.sig = getattr(fun, '__signature__', None)
    data.async_mode = async_mode
    if is_overload:
        data.overloads = ()
    else:
        data.overloads = tuple(describe_real_function(o, is_overload=True) for o in retrieve_overloads(fun))

    return data


################################################################################

def get_fun_defaults(fun: FunctionT, opt_args: strtup, flag: Literal['all', 'atoms', None]) -> dict[str, Any]:
    if not opt_args or flag is None:
        return {}
    defaults = getattr(fun, '__defaults__', None)
    if not defaults or type(defaults) is not tuple:
        return {}
    if flag == 'atoms':
        return {
            k: v for k, v in zip(opt_args, defaults)
            if is_atom_t(v) or is_class_t(v) or is_function_t(v)
        }
    else:
        return dict(zip(opt_args, defaults))


################################################################################

def create_virtual_function(data: FunctionData, handle: ResourceHandle) -> FunctionT:

    name, qname = data.name, data.qname
    qname = qname or name
    lineno = data.lineno
    annos = data.annos
    arg_defaults = data.defaults

    handle.name = name
    handle.kind = Kind.Function
    handle.keys = list(data.keys)
    handle.open = False

    is_async = data.is_async
    async_mode = data.async_mode

    if async_mode not in ('sync', None):
        assert is_async

    if is_async and async_mode == 'sync':
        async_mode = 'coro'

    if is_async and async_mode is None:
        async_mode = flags.DEFAULT_ASYNC_MODE

    if async_mode == 'future':
        import asyncio
        if 'return' in annos:
            annos['return'] = asyncio.Future[annos['return']]
        else:
            # https://peps.python.org/pep-0484/#the-meaning-of-annotations
            annos['return'] = asyncio.Future[Any]

    meta = {
        VHDL: handle,
        NAME: name,
        QUALNAME: qname,
        MODULE: data.module,
        DOC: data.doc,
        ANNOS: annos,
    }

    prm_str, pos_str, key_str = make_template_strs(data)

    modifiers = ''
    if async_mode == 'future':
        modifiers += ".set_async_mode('future')"
    elif async_mode == 'coro':
        modifiers += ".set_async_mode('coro')"

    result = f'''return {VHDL_ARG}.hdlr({VHDL_ARG}, {CALL_ARG}({INNER}, {pos_str}, {key_str}){modifiers})'''

    # this prevents us from virtualizing 'init', which has already happened because we
    # called '__new__' via RPC
    if qname.endswith('.__init__'):
        result = 'pass'

    code = f'''
def {MAKER}({VHDL_ARG}, {CALL_ARG}, {ARG_DEFAULTS}, {BLANK_DEFAULT}):
    def {INNER}{prm_str}:
        {result}
    return {INNER}'''

    try:
        exec(code, ns := {})
    except BaseException as err:
        fst = code.strip().splitlines()[1].strip()
        raise E.WarpDecodingError(
            f'function stub error!\n{handle=!r}\nsig={fst!r}\n{data=!r}\n{err=!r}\n\n{code}'
        )

    v_fun_maker = ns[MAKER]
    v_fun = v_fun_maker(handle, ResourceCallFunction, arg_defaults, ARG_DEFAULT)
    assert isinstance(v_fun, FunctionType)
    multi_set_raw(v_fun, meta)
    v_fun.__code__ = v_fun.__code__.replace(co_name=name, co_qualname=qname, co_firstlineno=lineno)

    if sig := data.sig:
        v_fun.__signature__ = sig

    if async_mode == 'coro':
        from inspect import markcoroutinefunction
        markcoroutinefunction(v_fun)

    for overload in data.overloads:
        overload_fn = create_overload_dummy_function(overload)
        register_overload(v_fun, overload_fn)

    return v_fun

################################################################################

def create_proxy_function(data: FunctionData, impl_fn: Callable) -> FunctionType:

    name = data.name
    qname = data.qname or name
    lineno = data.lineno
    prm_str, pos_str, key_str = make_template_strs(data)

    fwd_str = combine_pos_key_str(pos_str, key_str)
    result = f'''return {IMPL_FN_ARG}({fwd_str})'''

    arg_defaults = data.defaults if data.defaults else {}
    code = f'''
def {P_MAKER}({IMPL_FN_ARG}, {ARG_DEFAULTS}, {BLANK_DEFAULT}):
    def {P_INNER}{prm_str}:
        {result}
    return {P_INNER}'''

    exec(code, ns := {})

    p_fun_maker = ns[P_MAKER]
    p_fun = p_fun_maker(impl_fn, arg_defaults, ARG_DEFAULT)
    assert isinstance(p_fun, FunctionType)
    p_fun.__name__ = name
    p_fun.__qualname__ = qname
    p_fun.__module__ = data.module
    p_fun.__code__ = p_fun.__code__.replace(co_name=name, co_qualname=qname, co_firstlineno=lineno)
    p_fun.__doc__ = data.doc
    p_fun.__annotations__ = data.annos
    # TODO: set __text_signature__ to something helpful here...

    if sig := data.sig:
        p_fun.__signature__ = sig

    if data.is_async:
        from inspect import markcoroutinefunction
        markcoroutinefunction(p_fun)

    for overload in data.overloads:
        overload_fn = create_overload_dummy_function(overload)
        register_overload(p_fun, overload_fn)

    return p_fun

################################################################################

def _overload_dummy(*args, **kwds):
    """Helper for @overload to raise when called."""
    del args  # for vulture
    del kwds
    raise NotImplementedError(
        "You should not call an overloaded function. "
        "A series of @overload-decorated functions "
        "outside a stub module should always be followed "
        "by an implementation that is not @overload-ed.")

def create_overload_dummy_function(data: FunctionData) -> FunctionType:
    import typing
    dummy = getattr(typing, '_overload_dummy', None)
    if dummy is None:
        dummy = _overload_dummy
    o_fun = create_proxy_function(data, dummy)
    return o_fun

################################################################################

# these have to be not clash with func arg names, so we use constants to make the above code more
# readable

ARG_DEFAULTS = '___ARG_DEFAULTS___'
BLANK_DEFAULT = '___BLANK_DEFAULT___'

MAKER = '___MAKE_VIRTUAL_FN___'
INNER = '___VIRTUAL_FN___'

P_MAKER = '___MAKE_PROXY_FN___'
P_INNER = '___PROXY_FN___'

V_CALLBACK_ARG = '___VCALLBACK_ARG___'
GRID_ARG = '___GRID_ARG___'
CALL_ARG = '___CALL_ARG___'
HANDLER_ARG = '___HANDLER_ARG___'
VHDL_ARG = '___VHDL_ARG___'
IMPL_FN_ARG = '___IMPL_FN_ARG___'

################################################################################

def make_template_strs(data: FunctionData) -> tuple[str, str, str]:
    """
    This function returns a triple (prm_str, pos_str, key_str):

    1) param_str: a string like '(arg1, /, *args, **kwargs)', specifying the function
    arguments as they should appear in the `def NAME{prm_str}`.

    2) pos_str: a string like '(arg1, *args)', specifying the positional arguments
    as they should be forwarded as a tuple.

    3) key_str: a string like '{key1=val1, **kwargs}', specifying the keyword
    arguments as thy should be forwarded as a dictionary.

    You can build the 'combined' arguments as they should be forwarded to the
    *true* implementation via `combine_pos_key_str(pos_str, key_str)`.
    """
    reg_args = data.args
    pos_args = data.pos_args
    key_args = data.key_args
    pos_star = data.pos_star
    key_star = data.key_star
    opt_args = data.opt_args
    arg_defs = data.defaults

    prm_strs, add_prm = mklist()
    pos_strs, add_pos = mklist()
    key_strs, add_key = mklist()

    def add_arg(name: str):
        if name in opt_args:
            if name in arg_defs:
                add_prm(f'{name}={ARG_DEFAULTS}[{name!r}]')
            else:
                add_prm(f'{name}={BLANK_DEFAULT}')
        else:
            add_prm(name)

    if pos_args:
        for arg in pos_args:
            add_arg(arg)
            add_pos(arg)
        add_prm('/')

    for arg in reg_args:
        add_arg(arg)
        if arg in opt_args:
            add_key(f'{arg!r}:{arg}')
        else:
            add_pos(arg)

    if pos_star:
        add_arg('*' + pos_star)

    if key_args:
        if not pos_star:
            add_prm('*')
        for arg in key_args:
            add_arg(arg)
            add_key(f'{arg!r}:{arg}')

    if key_star:
        add_arg('**' + key_star)

    if len(pos_strs) == 1:
        pos_strs.append('')

    prm_str = '(' + commas(prm_strs) + ')'
    pos_str = '(' + commas(pos_strs) + ')'
    key_str = '{' + commas(key_strs) + '}'

    if key_star:
        if key_str == '{}':
            key_str = key_star
        else:
            key_str = key_str[1:-1].removesuffix(', ')
            key_str = f'{{{key_str}, **{key_star}}}'

    if pos_star:
        if pos_str == '()':
            pos_str = pos_star
        else:
            pos_str = pos_str[1:-1].removesuffix(', ')
            pos_str = f'({pos_str}, *{pos_star})'

    return prm_str, pos_str, key_str


commas = ', '.join

################################################################################

def combine_pos_key_str(pos_str, key_str):
    if pos_str.startswith('(') and pos_str.endswith(')'):
        pos_str = pos_str[1:-1].rstrip(', ')
    else:
        # if pos_str is just 'args'
        pos_str = f'*{pos_str}'

    if key_str:
        key_str = f'**{key_str}'

    if not pos_str and not key_str:
        return ''
    if not pos_str:
        return key_str
    if not key_str:
        return pos_str
    return f'{pos_str}, {key_str}'

################################################################################

# attach definitions to class
FunctionData.describe_resource = staticmethod(describe_real_function)
FunctionData.create_resource = create_virtual_function



# ruff: noqa
# fmt: off

import pytest

from agentica_internal.testing.warpc.virtual_data_tests import verify_plain_old_data


@pytest.mark.asyncio
async def test_warp_plain_old_data():
    await verify_plain_old_data()



# fmt: off

from .__ import *
from .handle import ResourceHandle, get_handle

__all__ = [
    'ResourceData',
    'ResourceHandle',
    'get_handle',
]

################################################################################

#
# SNIPPETS = {
#     'ClassMsg':           'enc.enc_cls(data.{K})            data.{K} = dec.dec_cls(self.{K})',
#     'FunctionMsg':        'enc.enc_fun(data.{K})            data.{K} = dec.dec_fun(self.{K})',
#     'ValueT':             'enc.enc_val(data.{K})            data.{K} = dec.dec_val(self.{K})',
#     'MethodsMsg': 'mapdict(enc.enc_fun, data.{K})   data.{K} = mapdict(dec.dec_fun, '
#                           'self.{K})',
#     Rec[ResourceT]: 'mapdict(enc.enc_res, data.{K})   data.{K} = mapdict(dec.dec_fun, self.{K})',
#     AnnotationsT:   'enc.enc_ann(data.{K})            data.{K} = dec.dec_ann(self.{K})',
#
# }
#
#     'Tup[ClassT]': CLS_TUP = 1
#     'Tup[ClassT]': FUN_REC = 2
#     'Tup[ClassT]': TYP_REC
#     'Tup[ClassT]': VAL_REC
#     'Tup[ClassT]': SEQ
#     'Tup[ClassT]': VAL
#     'Tup[ClassT]': TYP_REC
#     'Tup[ClassT]': VAL_OR_NONE
#     'Tup[ClassT]': RES_REC
#     'Tup[ClassT]': SEQ_OR_VAL
#
# def compile_data_to_msg_encoder(attr: str, anno):
#     if cls(anno) is ForwardRef:
#
#

################################################################################

class ResourceData:
    __slots__ = ()

    STR_NAME: ClassVar[str]
    FORBIDDEN_FORM: ClassVar[ResourceT] = forbidden_object
    MSG_CLS: ClassVar['type[ResourceDataMsg]']
    MIGHT_ALIAS: ClassVar[bool] = False

    def encode(self, enc: EncoderP) -> 'ResourceDataMsg':
        cls = type(self)
        msg_cls = cls.MSG_CLS
        kwargs = msg_cls.encode_fields(self, enc)
        for key in msg_cls.__struct_fields__:
            if key not in kwargs:
                kwargs[key] = getattr(self, key)
        msg = msg_cls(**kwargs)
        return msg

    def repr(self):
        cls = type(self)
        if hasattr(cls, '__slots__'):
            dct = {}
            for slot in cls.__slots__:
                dct[slot] = getattr(self, slot, FIELD_ABSENT)
        else:
            dct = self.__dict__
        f_args = ', '.join(f'{k}={f_slot(k, v)}' for k, v in dct.items())
        return f'{cls.__name__}({f_args})'

    __str__ = repr
    __repr__ = repr

    def short_str(self) -> str:
        cls_name = type(self).__name__
        name = getattr(self, 'name', '...')
        return f'{cls_name}({name})'

    def pprint(self, err: bool = False):
        from ...core.fmt import f_slot_obj
        P.rprint(f_slot_obj(self), err=err)

    @classmethod
    def describe_resource(cls, resource: ResourceT, /) -> 'ResourceData': ...

    def create_resource(self, handle: 'ResourceHandle', /) -> ResourceT: ...

    @classmethod
    def forbidden_msg(cls) -> 'SystemResourceMsg':
        from ..msg.term_resource import SystemResourceMsg
        from ..system import LRID_TO_SRID
        forb_cls = cls.FORBIDDEN_FORM
        forb_sid = LRID_TO_SRID[id(forb_cls)]
        return SystemResourceMsg(forb_sid)

    def rename(self, tmpl: str | None, grid: GlobalRID) -> None:
        if tmpl and hasattr(self, 'name'):
            f_kind = type(self).__name__.removesuffix('Data')
            wid, fid, lid = grid
            name = tmpl.format(rsrc_name=self.name, rsrc_kind=f_kind, wid=wid, fid=fid, lid=lid)
            setattr(self, 'name', name)
            if type(getattr(self, 'qname', None)) is str:
                setattr(self, 'qname', name)



# fmt: off

__all__ = [
    'unbound_instance_method',
    'bound_instance_method',
    'bound_class_method',
    'builtin_bound_class_method',
    'builtin_bound_instance_method',
    'builtin_bound_dunder_method',
    'builtin_bound_dunder_method',
    'METHODS',
]

class Class:
    """Class docstring"""

    def instance_method(self, inst_method_arg: object) -> str:
        """Class.inst_method docstring"""
        ...

    @staticmethod
    def static_method(static_method_arg: bytes) -> object:
        """Class.static_method docstring"""
        ...

    @classmethod
    def class_method(cls, class_method_arg: int) -> int:
        """Class.class_method docstring"""
        ...

obj = Class()

unbound_instance_method = Class.instance_method
bound_instance_method = obj.instance_method
bound_class_method = Class.class_method
builtin_bound_class_method = bytes.hex
builtin_bound_instance_method = bytes().__buffer__
builtin_unbound_dunder_method = Class.__str__
builtin_bound_dunder_method = obj.__str__

METHODS = [
    unbound_instance_method,
    bound_instance_method,
    bound_class_method,
    builtin_bound_class_method,
    builtin_bound_instance_method,
    builtin_unbound_dunder_method,
    builtin_bound_dunder_method,
]



# ruff: noqa
# fmt: off

import pytest

from agentica_internal.testing.examples.annos import TreeAlias
from agentica_internal.testing.warpc.virtual_enum_tests import (
    ENUM_CLASSES,
    ENUM_INSTANCES,
    verify_enum_cls,
    verify_enum_obj,
)


@pytest.mark.asyncio
@pytest.mark.parametrize('enum_cls', ENUM_CLASSES)
async def test_warp_enum_class(enum_cls):
    await verify_enum_cls(enum_cls)


@pytest.mark.asyncio
@pytest.mark.parametrize('obj', ENUM_INSTANCES)
async def test_warp_enum_object(obj):
    await verify_enum_obj(obj)



# fmt: off

from dataclasses import KW_ONLY, InitVar, dataclass, field
from typing import ClassVar

from ..shared import *

__all__ = [
    'DClass',
    'DClassSub',
    'DClassFields',
    'DClassFrozen',
    'DClassVars',
    'DClassKWONLY',
    'DCLASS_CLASSES',
    'DCLASS_INSTANCES',
]


@dataclass
class DClass:
    a: int
    b: bool

    def __post_init__(self):
        pass

    def my_dc_method(self, z: str):
        pass


@dataclass
class DClassSub(DClass):
    c: str


@dataclass(frozen=True)
class DClassFields:
    a: int = field(default=_a, kw_only=False)
    b: bool = field(default=_b, kw_only=False)
    c: str = field(default=_c, kw_only=True)


@dataclass(frozen=True)
class DClassFrozen:
    a: int
    b: bool
    c: str


@dataclass
class DClassKWONLY:
    a: int
    b: bool
    _: KW_ONLY
    c: str


@dataclass
class DClassVars:
    a: int
    b: bool
    c: str

    cv: ClassVar[int]
    iv: InitVar[int]


DCLASS_CLASSES: list[type] = [
    DClass,
    DClassSub,
    DClassFields,
    DClassFrozen,
    DClassVars,
    DClassKWONLY,
]

DCLASS_INSTANCES: list[object] = [
    DClass(_a, _b),
    DClassSub(_a, _b, _c),
    DClassFields(_a, _b, c=_c),
    DClassKWONLY(_a, _b, c=_c),
    DClassVars(_a, _b, _c, iv=99),
]



from typing import Any

from agentica_internal.repl.repl import *


def verify_print_hook():
    class MaklovichRepl(BaseRepl):
        # https://www.youtube.com/watch?v=lIpev8JXJHQ
        def fmt_print_arg(self, value: Any, /) -> str:
            return 'Malkovich'

    repl = MaklovichRepl()
    repl_eval = repl.run_code("print(1, 2, 3)")
    assert repl_eval.output == 'Malkovich Malkovich Malkovich\n'


def verify_display_hook():
    class AngryRepl(BaseRepl):
        def fmt_display_arg(self, value: Any, /) -> str:
            return repr(value).upper()

    repl = AngryRepl()
    repl_eval = repl.run_code("'ok'")
    assert repl_eval.out_value == 'ok'
    assert repl_eval.out_str == "'OK'"
    assert repl_eval.output == "'OK'\n"

    repl_eval = repl.run_code("def foo(): pass")
    assert not repl_eval.has_out_value
    assert repl_eval.out_str is None


def verify_exception_hook():
    exceptions = set()

    class MyRepl(BaseRepl):
        def on_uncaught_exception(self, evaluation, error, /) -> None:
            exceptions.add(type(error.exception))

    repl = MyRepl()
    repl.run_code("1 / 0")
    repl.run_code("foo")
    repl.run_code("{}[1]")
    assert exceptions == {ZeroDivisionError, NameError, KeyError}


def verify_import_hook():
    imports = []

    class MyRepl(BaseRepl):
        def on_import(self, evaluation, name, items) -> None:
            imports.append((name, items))

    repl = MyRepl()
    ev = repl.run_code('''
    import builtins
    import time
    from math import sqrt
    ''')
    assert ev.error is None
    assert ev.output == ''
    assert ev.imported == ['builtins', 'time', 'math']
    assert imports == [
        ('builtins', None),
        ('time', None),
        ('math', ('sqrt',)),
    ]


if __name__ == '__main__':
    verify_print_hook()
    verify_display_hook()
    verify_exception_hook()
    verify_import_hook()



# fmt: off

import asyncio
import re

from agentica_internal.warpc.worlds.debug_world import *


async def verify_virtual_queue():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        queue = asyncio.Queue()
        queue_b = await B(queue)
        assert queue_b is not queue

        assert is_virtual(queue_b)
        queue_a = await A(queue_b)
        assert queue_a is queue

        assert type(queue_b).__name__ == 'Queue'
        assert type(queue_b) is not type(queue)

        assert queue.qsize() == 0
        assert queue_b.qsize() == 0
        queue_b.put_nowait(55)

        assert queue.qsize() == 1
        assert queue_b.qsize() == 1

        assert queue.get_nowait() == 55


async def verify_regex_patt():

    pair = DebugWorld.connected_pair(logging=True, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        patt = re.compile('.')
        patt_b = await B(patt)
        assert patt_b is not patt
        assert is_real(type(patt_b)) and is_real(patt_b)
        assert type(patt_b) is type(patt)

        patt_a = await A(patt_b)
        assert patt_a is patt_b

        assert patt_a == patt_b


async def verify_regex_match():

    pair = DebugWorld.connected_pair(logging=True, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        patt = re.compile('A+')
        match = patt.search(' AAA ')
        match_b = await B(match)
        assert match_b is not match
        assert is_real(type(match_b)) and is_real(match_b)
        assert type(match_b) is type(match)

        match_a = await A(match_b)
        assert repr(match_a) == repr(match_b)


if __name__ == '__main__':
    asyncio.run(verify_virtual_queue())
    asyncio.run(verify_regex_patt())
    asyncio.run(verify_regex_match())



# ruff: noqa
# fmt: off

import pytest
from agentica_internal.testing.warpc.func_default_tests import *


@pytest.mark.asyncio
async def test_no_func_defaults():
    await verify_no_func_defaults()

@pytest.mark.asyncio
async def test_atom_func_defaults():
    await verify_atom_func_defaults()

@pytest.mark.asyncio
async def test_all_func_defaults():
    await verify_all_func_defaults()



# fmt: off

from .__ import *
from .base import Msg

__all__ = [
    'VarsMsg',
    'NO_VARS'
]


################################################################################

if TYPE_CHECKING:
    from .resource_def import DefinitionMsg
    from .term import TermMsg
    from ..request.vars import Vars

################################################################################

class VarsMsg(Msg):

    vars: 'Rec[TermMsg]' = {}
    defs: 'Tup[DefinitionMsg]' = ()

    @staticmethod
    def encode(enc: EncoderP, vars_: Rec[TermT], fmt: EncodeFmt) -> 'VarsMsg':
        if fmt == 'full':
            enc_ctx = enc.enc_context()
            with enc_ctx:
                rec = enc.enc_record(vars_)
            defs = enc_ctx.enc_context_defs()
        elif fmt == 'json':
            rec = {k: enc_json(v) for k, v in vars_.items()}
            defs = ()
        else:
            raise ValueError(f'unsupported fmt: {fmt!r}')
        return VarsMsg(rec, defs)

    def decode(self, dec: DecoderP) -> 'Vars':
        from ..request.vars import Vars
        dec_ctx = dec.dec_context(self.defs)
        with dec_ctx:
            vars_ = dec.dec_record(self.vars)
        return Vars(vars_)


NO_VARS = VarsMsg()



"""Presumed unreachable code paths should raise errors which the user can use to report bugs."""

from typing import Any

from . import base

GITHUB_ISSUES_URL = 'https://github.com/symbolca-ai/agentica-issues/issues'


class ThisIsABug(base.AgenticaError):
    """Raise this error when reaching this path in the code should be considered a bug."""

    def __init__(self, *args: Any, report_url: str | None = None, dump_path: str | None = None):
        super().__init__(f'This should not have happened: {', '.join(map(str, args))}')
        if report_url is None:
            report_url = GITHUB_ISSUES_URL
        super().add_note(f'Please file a bug report at {report_url}')
        if dump_path is not None:
            # FUTURE: auto-generate these here.
            super().add_note('Please include your crash dump: {dump_path}')



# fmt: off

from agentica_internal.cpython.classes.anno import TForward

from .__ import *
from .term import *

__all__ = [
    'AtomMsg',
    'NumberMsg',
    'StrLikeMsg',
    'StrMsg',
    'BytesMsg',
    'SingletonMsg',
    'NoneMsg',
    'NotImplMsg',
    'EllipsisMsg',
    'ClosedMsg',
]

class ForwardRefTypeMsg(TermPassByValMsg, tag='fwdref'):
    type V = str
    v: str

    def decode(self, dec: DecoderP) -> V:
        return TForward(self.v)

    @classmethod
    def encode_atom(cls, term: V) -> 'ForwardRefTypeMsg':
        return cls(v=term)


################################################################################

class AtomMsg(TermPassByValMsg):
    """ABC for messages describing immutable, atomic values."""

    type V = AtomT

    def decode(self, fn: DecoderP) -> V: ...

    def decode_atom(self) -> V: ...

    @classmethod
    def encode_atom(cls, term: V) -> 'AtomMsg':
        if isinstance(term, (int, bool, float)):
            return NumberMsg.encode_atom(term)
        if isinstance(term, str):
            return StrMsg.encode_atom(term)
        if isinstance(term, bytes):
            return BytesMsg.encode_atom(term)
        if term is None:
            return NoneMsg.MSG
        if term is Ellipsis:
            return EllipsisMsg.MSG
        if term is NotImplemented:
            return NotImplMsg.MSG
        raise E.WarpEncodingError(f"not an atom: {term}")

################################################################################

class NumberMsg(AtomMsg, tag='num'):

    type V = NumberT
    v:       NumberT

    @property
    def shape(self) -> str:
        val = self.v
        cls = type(val)
        if cls is int:
            return str(val) if -9 <= val <= 9 else 'int'
        if cls is bool:
            return 'true' if val else 'false'
        return 'float'

    def decode(self, fn: DecoderP) -> V:
        return self.v

    def decode_atom(self) -> V:
        return self.v

    @classmethod
    def encode_atom(cls, term: V) -> 'NumberMsg':
        return cls(term)


################################################################################

class StrLikeMsg(AtomMsg):
    """ABC for messages describing `str` or `bytes` values."""

    type V = StrLikeT

    v: StrLikeT

    @property
    def shape(self) -> str:
        s = type(self).TAG
        v = self.v
        if len(v) <= 5:
            return repr(v)
        else:
            return s

    def decode(self, fn: DecoderP) -> V:
        return self.v

    def decode_atom(self) -> V:
        return self.v

    @classmethod
    def encode_atom(cls, term: V) -> 'StrLikeMsg':
        return cls(term)


################################################################################

class StrMsg(StrLikeMsg, tag='str'):

    type V = str
    v: str


class BytesMsg(StrLikeMsg, tag='bytes'):
    type V = bytes
    v: bytes


################################################################################

class SingletonMsg(AtomMsg):
    """ABC for messages representing singleton values like `None`. They have
    no content, since there is only one possible value they describe."""

    type V = SingletonT

    VAL: ClassVar[SingletonT]
    MSG: ClassVar['SingletonMsg']

    def decode(self, fn: DecoderP) -> V:
        return type(self).VAL

    def decode_atom(self) -> V:
        return type(self).VAL

    @classmethod
    def encode_constant(cls):
        return cls.MSG


################################################################################

class NoneMsg(SingletonMsg, tag='none'):
    type V = S.NoneT

    VAL = None


class NotImplMsg(SingletonMsg, tag='notimpl'):
    type V = S.NotImplT

    VAL = NotImplemented


class EllipsisMsg(SingletonMsg, tag='ellip'):
    type V = S.EllipT

    VAL = Ellipsis


class ClosedMsg(SingletonMsg, tag='closed'):
    type V = CLOSED

    VAL = CLOSED


################################################################################

NoneMsg.MSG      = NoneMsg()
NotImplMsg.MSG   = NotImplMsg()
EllipsisMsg.MSG  = EllipsisMsg()
ClosedMsg.MSG    = ClosedMsg()



# fmt: off

from ...core.sentinels import SENTINELS
from .__ import *
from .base import Msg

__all__ = [
    'ResultMsg',
    'ResultValueMsg',
    'OkMsg',
    'ValueMsg',
    'JsonValueMsg',
    'ResultErrorMsg',
    'ErrorMsg',
    'InternalErrorMsg',
    'ResultUnavailableMsg',
    'OK_MSG',
]


################################################################################

if TYPE_CHECKING:
    from .term import TermMsg
    from .term_exception import ExceptionMsg

################################################################################

class ResultMsg(Msg):
    """
    ABC for messages describing `FramedResponseMsg` content; these decode to `Result`.

    `ResultValueMsg` and `ResultErrorMsg` are sub-ABCs for representing
    `Result.good(...)` and `Result.bad(...)` respectively.
    """
    type V = Result

    is_ok:  ClassVar[bool]
    is_err: ClassVar[bool]

    def decode(self, dec: DecoderP, /) -> Result:
        raise NotImplementedError(type(self))

    def decode_json(self) -> Result:
        raise NotImplementedError(type(self))

    @staticmethod
    def encode(enc: EncoderP, res: Result, fmt: DecodeFmt = 'full', /) -> 'ResultMsg':
        if res.is_ok:
            value = res.value
            if fmt == 'full':
                if res is OK_RESULT:
                    return OK_MSG
                term_msg = enc.enc_any(value)
                return ValueMsg(term_msg)
            elif fmt == 'json':
                json_data = enc_json(value)
                return JsonValueMsg(json_data)
            elif fmt == 'type':
                from .term_atom import StrMsg
                cls = type(value)
                cls_name = cls.__qualname__
                if not is_str(cls_name):
                    cls_name = cls.__name__
                if is_str(cls_module := cls.__module__):
                    cls_name = f'{cls_module}.{cls_name}'
                return ValueMsg(StrMsg(cls_name))
            elif fmt == 'schema':
                json_schema = get_json_schema(value)
                json_data = enc_json(json_schema)
                return JsonValueMsg(json_data)
            elif fmt == 'raw':
                term_msg = enc.enc_any(value)
                term_data = term_msg.to_msgpack()
                return RawValueMsg(term_data)
            else:
                return InternalErrorMsg(f'invalid message format: {fmt}')
        elif res.is_err:
            error = res.error
            if isinstance(error, E.WarpError):
                error_str = str(error)
                return InternalErrorMsg(error_str)
            elif isinstance(error, BaseException):
                error_msg = enc.enc_exception(error)
                return ErrorMsg(error_msg)
            else:
                return InternalErrorMsg('invalid exception error')
        else:
            sentinel: Sentinel = res.value
            reason = sentinel.name if isinstance(sentinel, Sentinel) else None
            return ResultUnavailableMsg(reason)


################################################################################

class ResultValueMsg(ResultMsg):
    """ABC for messages decoding to `Result.good`."""

    is_ok:  ClassVar[bool] = True
    is_err: ClassVar[bool] = False

    @staticmethod
    def encode_value(enc: EncoderP, val: TermT) -> 'JsonValueMsg': ...


################################################################################

class OkMsg(ResultValueMsg, tag='ok'):
    """Message that just means 'success', conventionally `None` in Python."""

    @staticmethod
    def encode_value(enc: EncoderP, val: TermT) -> 'OkMsg':
        return OK_MSG

    def decode(self, dec: DecoderP) -> Result:
        return OK_RESULT


################################################################################

class ValueMsg(ResultValueMsg, tag='val'):
    """Message that contains some arbitrary value via `TermMsg`."""

    val: 'TermMsg'

    def decode(self, dec: DecoderP) -> Result:
        return Result.good(self.val.decode(dec))

    def __shape__(self) -> str:
        return self.val.shape


################################################################################

class JsonValueMsg(ResultValueMsg, tag='json'):
    """Message that describes a pure JSON-like value."""

    json: bytes

    def decode(self, dec: DecoderP) -> Result:
        value = dec_json(self.json)
        return Result.good(value)

    def __shape__(self) -> str:
        return f'..{len(self.json)}'



################################################################################

class RawValueMsg(ResultValueMsg, tag='raw'):
    """Message that contains the raw bytes of an actual term message."""

    data: bytes

    def decode(self, dec: DecoderP) -> Result:
        return Result.good(self.data)

    def __shape__(self) -> str:
        return f'..{len(self.data)}'

################################################################################

class ResultErrorMsg(ResultMsg):
    """ABC for messages decoding to `Result.bad(...)`."""

    is_ok:  ClassVar[bool] = False
    is_err: ClassVar[bool] = True

################################################################################

class ErrorMsg(ResultErrorMsg, tag='err'):
    """Message encoding a caught exception, encoded as `ExceptionMsg`."""

    exc: 'ExceptionMsg'

    def __shape__(self) -> str:
        return self.exc.shape

    def decode(self, dec: DecoderP) -> Result:
        exception = dec.dec_exception(self.exc)
        return Result.bad(exception)


################################################################################

class InternalErrorMsg(ResultErrorMsg, tag='interr'):
    """Message encoding an internal exception or error condition, which will
    become a `Result.bad(RuntimeError(...))`.
    """

    error: str

    def __shape__(self) -> str:
        return self.error

    def decode(self, dec: DecoderP) -> Result:
        exception = RuntimeError(f'Internal error: {self.error}')
        return Result.bad(exception)


################################################################################

class ResultUnavailableMsg(ResultMsg):

    reason: str | None = None

    is_ok:  ClassVar[bool] = False
    is_err: ClassVar[bool] = False

    def decode(self, dec: DecoderP) -> Result:
        sentinel = SENTINELS.get(self.reason)
        return Result.unavailable(sentinel)


################################################################################

OK_MSG = OkMsg()



# fmt: off

import asyncio
from typing import *

from agentica_internal.core.type import anno_str
from agentica_internal.testing import *
from agentica_internal.testing.examples import ANNOS, CLASSES
from agentica_internal.testing.examples.annos import TreeAlias
from agentica_internal.testing.examples.classes import ENUM_CLASSES
from agentica_internal.warpc.worlds.debug_world import DebugWorld

SKIP = (
    'VarInt',  # we don't serialize TypeVar bounds yet
)

ANNOS = [a for a in ANNOS if a not in ENUM_CLASSES]
CLASSES = [a for a in CLASSES if a not in ENUM_CLASSES]

async def verify_annotation_equality(anno: Any):

    # verify that either deserialized annotation compares equal to original,
    # or it formats the same (because e.g. buried classes will be different)
    pair = DebugWorld.connected_pair(qualify_names=False)

    async with pair as B:
        if anno is TreeAlias:
            return pair
        anno_a = anno
        anno_b = await B(anno)
        str_a = anno_str(anno_a)
        str_b = anno_str(anno_b)
        obj_equal = anno_a == anno_b
        str_equal = str_a == str_b
        if any(skip in str_a for skip in SKIP):
            return pair
        assert obj_equal or str_equal, f"annotations not equal: {str_a} == {str_b}"

    return pair


async def verify_class_annotation_equality(cls: type):

    # verify that either deserialized annotation compares equal to original,
    # or it formats the same (because e.g. buried classes will be different)
    pair = DebugWorld.connected_pair(qualify_names=False)

    async with pair as B:
        anno_a = cls.__annotations__
        cls_b = await B(cls)
        anno_b = cls_b.__annotations__
        str_a = anno_str(anno_a)
        str_b = anno_str(anno_b)
        obj_equal = anno_a == anno_b
        str_equal = str_a == str_b
        if any(skip in str_a for skip in SKIP):
            return
        assert obj_equal or str_equal, f"class annotations not equal: {str_a} == {str_b}"

    return pair


def verify_normalizing_unions():
    # verify that the weird normalizing behavior of union types don't cause problems

    from types import NoneType

    from agentica_internal.warpc.messages import DefinitionMsg, SystemResourceMsg
    from agentica_internal.warpc.pure import PURE_CODEC
    from agentica_internal.warpc.resources import TypeUnionData

    def encode_decode_union(sys: bool, alts: tuple):
        grid = 9, 9, 9
        union_data = TypeUnionData()
        union_data.alts = alts
        union_data.sys = False
        def_msg = DefinitionMsg(grid, union_data.encode(PURE_CODEC))

        world = DebugWorld(logging=False)
        frame = world.root
        resource = frame.decode_incoming_definition(grid, def_msg)
        assert id(resource) not in frame.remote_lrids
        # ensure no aliasing occurred
        assert isinstance(frame.enc_any(NoneType), SystemResourceMsg)
        assert isinstance(frame.enc_any(int), SystemResourceMsg)
        return resource

    assert encode_decode_union(False, (int, str, str,)) == Union[str, int]
    assert encode_decode_union(False, (int,)) == int
    assert encode_decode_union(False, (NoneType, NoneType,)) is NoneType
    assert encode_decode_union(False, (NoneType,)) is NoneType
    assert encode_decode_union(False, ()) is NoReturn

    assert encode_decode_union(True, (int, str, str,)) == int | str
    assert encode_decode_union(True, (int,)) == int
    assert encode_decode_union(True, (NoneType, NoneType,)) is NoneType
    assert encode_decode_union(True, (NoneType,)) is NoneType
    assert encode_decode_union(True, ()) is NoReturn


def verify_annotation_equality_sync(anno: Any):
    return asyncio.run(verify_annotation_equality(anno))


def verify_class_annotation_equality_sync(anno: Any):
    return asyncio.run(verify_class_annotation_equality(anno))


def verify_annotations():
    run_object_tests(verify_annotation_equality_sync, ANNOS, on_error='s')


def verify_class_annotations():
    run_object_tests(verify_class_annotation_equality_sync, CLASSES, on_error='s')



def collect_anno_events(events: list):
    for anno in ANNOS:
        verify_annotation_equality_sync(anno).collect_events(events)
    for cls in CLASSES:
        verify_class_annotation_equality_sync(cls).collect_events(events)


if __name__ == '__main__':
    verify_annotations()
    verify_normalizing_unions()
    verify_class_annotations()



# fmt: off

from msgspec import Struct
from typing import ClassVar
import enum

from .kinds import Kind as DefinitionKind


__all__ = [
    'LanguageID',
    'TransUID',
    'DefinitionKind',
    'TranscodablePythonMessage',
    'TranscodableJavascriptMessage'
]


################################################################################

type TransUID = tuple[LanguageID, int]

################################################################################

class LanguageID(enum.IntEnum):
    PYTHON     = 0
    JAVASCRIPT = 1


################################################################################

class TranscodableMsg(Struct):

    LANGUAGE: ClassVar[LanguageID] = LanguageID.JAVASCRIPT

    @property
    def definition_kind(self) -> DefinitionKind:
        return self.__trans_def_kind__()

    @property
    def tid(self) -> TransUID:
        return self.LANGUAGE, self.__trans_int_uid__()

    ############################################################################

    def __trans_def_kind__(self) -> DefinitionKind:
        raise TypeError(f"{type(self).__name__} is not a resource definition message.")

    def __trans_int_uid__(self) -> int:
        raise TypeError(f"{type(self).__name__} does not have a transcoding UID.")


################################################################################

class TranscodablePythonMessage(TranscodableMsg):

    LANGUAGE: ClassVar[LanguageID] = LanguageID.PYTHON


################################################################################

class TranscodableJavascriptMessage(TranscodableMsg):

    LANGUAGE: ClassVar[LanguageID] = LanguageID.JAVASCRIPT



# fmt: off

import enum
import datetime as DT
import dataclasses as DC
import itertools as IT
import inspect as INS
import re as RE
import warnings
import os
import copyreg

from ...core.anno import union_iter

from .__ import *
from .term import *
from .msg_aliases import *


__all__ = [
    'SlotObjMsg',
    'ReduceObjMsg',
    'RegexPatternMsg',
    'RegexMatchMsg',
    'ClassUnionMsg',
    'EnumMemberMsg',
    'EnumKeyMsg',
    'EnumValMsg'
]


################################################################################

if TYPE_CHECKING:
    from .term_resource import ResourceMsg, SystemResourceMsg

################################################################################

type SlotObjT = DC.Field | DC._DataclassParams | DC.InitVar  # type: ignore

SLOT_OBJ_TYPES: tuple[type, ...] = SlotObjT.__value__.__args__  # type: ignore


################################################################################

class SlotObjMsg(TermPassByValMsg, tag='slots'):
    """Message describing objects by-value via their slot contents. Used
    for instances of certain builtin system classes that we wish to serialize by-value, like
    dataclass Fields."""

    type V = SlotObjT

    cls: 'SystemResourceMsg'
    slots: 'Rec[TermMsg]'

    def __len__(self) -> int:
        return len(self.slots)

    def __shape__(self) -> str:
        return self.cls.sys_name

    def decode(self, dec: DecoderP) -> V:
        from .term_resource import SystemResourceMsg
        assert isinstance(self.cls, SystemResourceMsg)
        _cls = self.cls.sys_cls
        term = object.__new__(_cls)  # type: ignore
        slots = dec.dec_record(self.slots)
        for slot_key, slot_val in slots.items():
            setattr(term, slot_key, slot_val)
        return term

    @classmethod
    def encode_compound(cls, term: V, enc: EncoderP) -> 'SlotObjMsg':
        _cls = type(term)
        slot_keys = _cls.__slots__
        cls_msg = enc.enc_class(_cls)
        assert isinstance(cls_msg, SystemResourceMsg), f"trying to encode non-system cls {_cls}"
        slots, add_slot = mkdict()
        enc_any = enc.enc_any
        for slot_key in slot_keys:
            val = getattr(term, slot_key, FIELD_ABSENT)
            if val is not FIELD_ABSENT:
                add_slot(slot_key, enc_any(val))
        return SlotObjMsg(cls_msg, slots)


################################################################################

type ReducibleBuiltinsObjT = (
    partial | range | map | filter | zip
)

type ReducibleItertoolsObjT = (
    IT.count | IT.islice | IT.cycle | IT.repeat | IT.takewhile | IT.dropwhile |
    IT.zip_longest | IT.starmap
)

type ReducibleOSObjT = (
    os.stat_result
)

type ReducibleDatetimeObjT = (
    DT.date | DT.time | DT.datetime | DT.timedelta
)

type ReducibleInspectObjT = (
    INS.Signature | INS.Parameter | INS.BoundArguments
)

type ReducibleObjT = (
    ReducibleBuiltinsObjT
    | ReducibleItertoolsObjT
    | ReducibleInspectObjT
    | ReducibleOSObjT
    | ReducibleDatetimeObjT
)

REDUCIBLE_TYPES: tuple[type, ...] = tuple(union_iter(ReducibleObjT, type))

################################################################################

class ReduceObjMsg(TermPassByValMsg, tag='reduce'):
    """Message describing objects by-value via the result of `__reduce__`."""

    type V = ReducibleObjT

    cls: 'SystemResourceMsg'
    args: 'Tup[TermMsg]'
    state: 'TermMsg | None'

    def __shape__(self) -> str:
        return self.cls.sys_name

    def decode(self, dec: DecoderP) -> V:
        assert isinstance(self.cls, SystemResourceMsg)
        _cls = self.cls.sys_resource  # might be a class, or 'iter', which is a function
        if _cls not in REDUCIBLE_TYPES:
            raise E.WarpDecodingError(f"{_cls!r} is not reducible")
        args = dec.dec_sequence(self.args)
        state = dec.dec_any(self.state) if self.state is not None else None
        try:
            obj = _cls(*args)
            obj.__setstate__(state) if state is not None else None
            return obj
        except BaseException as exc:
            raise E.WarpDecodingError(f"Could not expand {self}:\n{exc!r}")

    @classmethod
    def encode_compound(cls, term: V, enc: EncoderP) -> 'ReduceObjMsg':
        _cls = type(term)
        try:
            reduced = term.__reduce__()
            if len(reduced) == 2:
                clb, args = reduced
                assert clb is not copyreg._reconstructor
                state = None
            elif len(reduced) == 3:
                clb, args, state = reduced
            else:
                raise E.WarpEncodingError(f"bad reduction: {f_object_id(reduced)}")
            # assert clb is _cls, f"constructor is not original type: {clb} != {_cls}"
            assert type(args) is tuple, f"args is not a tuple: {f_object_id(args)}"
        except BaseException as exc:
            # f_exc = fmt_exception(exc)
            raise E.WarpEncodingError(f"Could not reduce {f_object_id(term)}: {exc}")
        cls_msg = enc.enc_system_resource(_cls)
        args_msg = enc.enc_sequence(args)
        state_msg = enc.enc_any(state)
        return ReduceObjMsg(cls_msg, args_msg, state_msg)


################################################################################

class RegexPatternMsg(TermPassByValMsg, tag='regex_pattern'):
    """Message describing re.Pattern objects by value."""

    type V = RE.Pattern

    pattern: str
    flags:   int

    def decode(self, dec: DecoderP) -> V:
        return RE._compile(self.pattern, self.flags)

    @classmethod
    def encode_atom(cls, term: V) -> 'RegexPatternMsg':
        return RegexPatternMsg(term.pattern, term.flags)


################################################################################

class RegexMatchMsg(TermPassByValMsg, tag='regex_match'):
    """Message describing re.Match objects by value."""

    type V = RE.Match

    pattern: str
    flags:   int
    string:  str
    span:   tuple[int, int]

    def decode(self, dec: DecoderP) -> V:
        # this is ugly but since re.Match is not pickle-able there is
        # no other way to do it!
        patt = RE._compile(self.pattern, self.flags)
        start, end = span = self.span
        for match in patt.finditer(self.string, start, end):
            if match.span() == span:
                return match
        raise E.WarpDecodingError(f"Could not reconstruct re.Match")

    @classmethod
    def encode_atom(cls, term: V) -> 'RegexMsg':
        re = term.re
        return RegexMatchMsg(re.pattern, int(re.flags), term.string, term.span())


################################################################################

class ClassUnionMsg(TermPassByValMsg, tag='class_union'):
    """Message for simple inline unions like `int | float` (used by typescript for 'Number')."""

    alts: 'ClassesTupleMsg'

    def decode(self, dec: DecoderP) -> TypeT:
        try:
            cls_list = []
            for msg in self.alts:
                cls = dec.dec_type(msg)
                if not isinstance(cls, type):
                    return Any
            union = cls_list[0]
            for cls in cls_list[1:]:
                union |= cls
            return union
        except:
            return Any


################################################################################

class EnumMemberMsg(TermPassByValMsg):

    cls: 'ResourceMsg'

    def decode(self, dec: DecoderP) -> enum.Enum: ...

    @staticmethod
    def encode_enum(value: enum.Enum, enc: EncoderP) -> 'EnumMemberMsg':
        cls_msg = enc.enc_class(type(value))
        if isinstance(value, int):
            # IntFlags in particular allow non-nameable enum values
            return EnumValMsg(cls_msg, enc.enc_any(int(value)))
        else:
            return EnumKeyMsg(cls_msg, value.name)


################################################################################

class EnumKeyMsg(EnumMemberMsg, tag='enum_key'):

    cls: 'ResourceMsg'
    key:  str

    def decode(self, dec: DecoderP) -> TypeT:
        enum_cls = dec.dec_class(self.cls)
        if not issubclass(enum_cls, enum.Enum):
            raise E.WarpEncodingError(f"{enum_cls=!r} is not an enum class")
        return enum_cls._member_map_[self.key]


################################################################################

class EnumValMsg(EnumMemberMsg, tag='enum_val'):

    cls: 'ResourceMsg'
    val: 'TermMsg'

    def decode(self, dec: DecoderP) -> TypeT:
        enum_cls = dec.dec_class(self.cls)
        enum_val = dec.dec_any(self.val)
        if not issubclass(enum_cls, enum.Enum):
            raise E.WarpEncodingError(f"{enum_cls=!r} is not an enum class")
        return enum_cls(enum_val)


################################################################################

# FIXME: replace this mechanism with virtual iterators
warnings.filterwarnings(
    'ignore',
    message='Pickle, copy, and deepcopy support will be removed from itertools in Python 3.14.',
    category=DeprecationWarning,
    append=True
)



from agentica_internal.core.fmt import (
    f_anno,
    f_atom,
    f_class,
    f_datum,
    f_object,
    f_object_id,
)
from agentica_internal.core.print import hprint, tprint

from .tst_utils.tst_run import run_object_tests, run_object_to_file_tests
from .tst_utils.tst_storage import FileTemplate

__all__ = [
    'run_object_tests',
    'run_object_to_file_tests',
    'FileTemplate',
    'f_anno',
    'f_object',
    'f_class',
    'f_atom',
    'f_object_id',
    'f_datum',
    'hprint',
    'tprint',
]



# fmt: off

from pathlib import Path
import os
from typing import NoReturn, BinaryIO
from collections.abc import Callable
from weakref import WeakSet

from ..fmt import f_id

from .log_flag import LogFlag
from .log_tags import *
from .log_fns import *
from .log_context import *
from ..strs import pathsafe_str

__all__ = [
    'LogBase',
    'binary_log_dir',
    'refresh_log_bases'
]

###############################################################################

class LogBase:
    __slots__ = 'logging', 'log_name', 'id_name', '_log1', '__weakref__'

    logging: bool
    log_name: str
    id_name: str

    def __init__(self, *, logging: bool | LogFlag = False, id_name: str = ''):
        id_name = id_name or f_id(id(self))
        self.logging = False
        if type(logging) is bool:
            logging = should_log_cls(logging, type(self))
        elif type(logging) is LogFlag:
            pass
        else:
            log_fn(type(self).__name__)('invalid keyword logging =', logging)
            logging = False
        self.logging = logging
        self.id_name = id_name
        self.log_name = f'{type(self).__name__}[{id_name}]'
        self._log1 = log_fn(self.log_name)
        self.log("__init__")

    def __short_str__(self) -> str:
        return self.log_name

    __repr__ = __str__ = __short_str__

    def refresh_logging_flag(self):
        self.logging = should_log_cls(self.logging, type(self))
        self._log1 = log_fn(self.log_name)

    def __refresh_when_tags_change__(self) -> None:
        REFRESHED_LOG_BASES.add(self)

    @property
    def name(self) -> str:
        return self.log_name

    def rename(self, name: str) -> None:
        self.log_name = name
        self._log1 = log_fn(name)

    @property
    def log(self) -> Callable[..., None]:
        return self._log1 if bool(self.logging) else no_log_fn

    @property
    def log_forced(self) -> Callable[..., None]:
        return self._log1

    def dump_exception(self, exception: BaseException, *args) -> None:
        return self._log1(*args, exception)

    def panic(self, *args) -> NoReturn:
        from ..print import panic
        panic(*args)

    @property
    def task_name(self) -> str:
        return ''.join(c for c in self.log_name if not c.islower())

    def log_as(self, label: str, *args) -> LogContext:
        return LogContext((self.log_name, label), *args) if self.logging else no_log_ctx

    def open_binary_log(self, suffix: str) -> BinaryIO | None:
        cls_name = type(self).__name__.upper()
        env_var = f'{cls_name}_BINARY_LOG_DIR'
        env_val = os.getenv(env_var)
        if env_val is None or '/' not in env_val:
            return None
        path = open_unique_binary_log(env_val, self.log_name, suffix)
        if not path:
            self._log1(f"binary log path '{path.as_posix()}' is invalid")
            return None
        self._log1(f"opened binary log to '{path.as_posix()}'")
        return path.open(mode='wb')

    @classmethod
    def enable_binary_log(cls, log_dir: str | Path = ''):
        cls_name = cls.__name__.upper()
        env_var = f'{cls_name}_BINARY_LOG_DIR'
        log_dir = binary_log_dir(log_dir)
        os.environ[env_var] = log_dir.as_posix()


def binary_log_dir(log_dir: str) -> Path:
    if not log_dir:
        log_dir = Path(os.getcwd(), 'logs')
    elif type(log_dir) is str:
        log_dir = Path(log_dir).absolute()
    else:
        assert type(log_dir) is Path
    if not log_dir.is_dir() and not log_dir.parent.is_dir():
        raise ValueError(f'{log_dir} is not a directory')
    return log_dir

###############################################################################

binary_log_count = 0

def open_unique_binary_log(dir_name: str, log_name: str, suffix: str) -> Path | None:
    global binary_log_count
    binary_log_count += 1
    name = pathsafe_str(log_name).replace('(', '_').replace(')', '_')
    pid = os.getpid()
    filename = f'{pid}_{name}_{binary_log_count-1:03}.{suffix}'
    filepath = Path(dir_name, filename)
    try:
        filepath.parent.mkdir(exist_ok=True)
        return filepath
    except:
        return None

###############################################################################

REFRESHED_LOG_BASES = WeakSet[LogBase]()

def refresh_log_bases():
    for log_base in REFRESHED_LOG_BASES:
        log_base.refresh_logging_flag()



import asyncio
from typing import Callable

from agentica_internal.warpc.worlds.all import DebugWorld


def bar(l: list) -> int:
    return sum(l)


def foo(l1: list, l2: list, other: Callable) -> list:
    return other(l1 + l2)


async def run():
    async with DebugWorld.connected_pair(logging='SEND+RECV') as B:
        b_fn = B(foo)
        b_call = b_fn([1, 2, 3], B([4, 5, 6]), bar)
        print(b_call, type(b_call))
        result = await b_call

        print(result)


asyncio.run(run(), debug=True)



# fmt: off

from ..__ import *
from ..attrs import VHDL
from .__raw import *

__all__ = [
    'is_virtual_object',
    'is_virtual_class',
    'ExecHandler',
    'ResourceHandle',
    'has_handle',
    'get_handle'
]


################################################################################

if TYPE_CHECKING:
    from ..request.request_resource import ResourceRequest

################################################################################

type ExecHandler = Callable[[ResourceHandle, 'ResourceRequest'], Any]

################################################################################

# ResourceHandle is buried inside a virtual resource (under the VHDL attribute)
# and contains enough information to resolve resource requests
class ResourceHandle:
    __slots__ = 'grid', 'fkey', 'hdlr', 'kind', 'keys', 'open', 'name', 'hash',

    # don't wish this to ever be visible
    if TYPE_CHECKING:
        grid: GlobalRID     # what is our global resource ID?
        fkey: FrameKey      # what frame carries information about us? CURRENTLY NOT USED
        hdlr: ExecHandler   # callback function to satisfy requests
        kind: Kind          # Kind.Class, Kind.Function, etc.
        keys: list[str]     # which `getattrs` should be let through
        open: bool          # True if there might be more keys we don't know about locally
        hash: int           # set the first time __hash__ is executed remotely
        name: str           # the unqualified name of the resource, if any

    def __repr__(self) -> str:
        return f"<{self.kind} {f_grid(self.grid)}>"

    __short_str__ = __str__ = __repr__

################################################################################

def has_handle(obj: Any) -> bool:
    try:
        if type(obj) is type or isinstance(obj, type):
            handle = cls_get(obj, VHDL)
        elif isinstance(obj, ModuleType):
            handle = mod_get(obj, VHDL)
        else:
            handle = obj_get(obj, VHDL)
        return type(handle) is ResourceHandle
    except:
        pass
    return False

def is_virtual_object(obj: Any) -> bool:
    try:
        handle = obj_get(obj, VHDL)
        return type(handle) is ResourceHandle
    except:
        pass
    return False

def is_virtual_class(cls: type) -> bool:
    try:
        if type(cls) is type or isinstance(cls, type):
            handle = cls_get(cls, VHDL)
            return type(handle) is ResourceHandle
    except:
        pass
    return False

def get_handle(obj: Any) -> ResourceHandle | None:
    try:
        if type(obj) is type or isinstance(obj, type):
            handle = cls_get(obj, VHDL)
        elif isinstance(obj, ModuleType):
            handle = mod_get(obj, VHDL)
        else:
            handle = obj_get(obj, VHDL)
        if type(handle) is ResourceHandle:
            return handle
    except:
         pass
    return None



# fmt: off

from collections.abc import Iterable
from enum import IntEnum
from typing import TypeVar, NamedTuple, Self, ClassVar, TYPE_CHECKING


__all__ = [
    'Tick',
    'Direction',
    'Incoming',
    'Outgoing',
    'EventType',
    'SimpleEvent',
    'Event',
    'FrameEvent',
    'FrameEnterEvent',
    'OutgoingFrameEnter',
    'IncomingFrameEnter',
    'FrameExitEvent',
    'OutgoingFrameExit',
    'IncomingFrameExit',
    'ResourceEvent',
    'OutgoingResource',
    'IncomingResource',
    'EventEvent',
    'IncomingEvent',
    'OutgoingEvent',
    'MessageEvent',
    'OutgoingMessage',
    'IncomingMessage',
]


################################################################################

if TYPE_CHECKING:
    from .msg.rpc import RPCMsg
    from .msg.rpc_event import EventMsg
    from .msg.rpc_framed import FramedRequestMsg, FramedResponseMsg
    from .resource.base import ResourceData

################################################################################

type Tick      = int
type EventType = type[Event]

class Direction(IntEnum):
    Incoming = 1
    Outgoing = 0

Incoming = Direction.Incoming
Outgoing = Direction.Outgoing

################################################################################

class SimpleEvent(NamedTuple):
    tick: Tick
    type: EventType

    def __repr__(self) -> str:
        return f'{self.tick:<3}\t{self.type.cls_name})'

    __str__ = __repr__

################################################################################

E = TypeVar('E', bound='Event')

class Event:
    __slots__ = 'tick',
    __icon__: ClassVar[str]

    INCOMING: ClassVar[EventType]
    OUTGOING: ClassVar[EventType]

    tick: Tick

    @property
    def cls_name(self) -> str:
        return type(self).__name__

    def short_str(self) -> str:
        short_args = ', '.join(self.__short_args__())
        return f"{self.cls_name}({self.tick}, {short_args})"

    def __str__(self) -> str:
        return self.short_str()

    def line_str(self) -> str:
        line_args = '\t'.join(self.__line_args__())
        return f"#{self.tick:<3}{line_args}"

    def __lt__(self, other: Self) -> bool:
        return self.tick < other.tick

    def __le__(self, other: Self) -> bool:
        return self.tick <= other.tick

    def __short_args__(self) -> Iterable[str]: ...

    def __line_args__(self) -> Iterable[str]: ...

    @classmethod
    def dir(cls: type[E], d: Direction) -> type[E]:
        return cls.INCOMING if d else cls.OUTGOING


################################################################################

class FrameEvent(Event):
    __slots__ = 'tick', 'request'

    tick:     Tick
    request:  'FramedRequestMsg'


################################################################################

class FrameEnterEvent(FrameEvent):
    __slots__ = 'tick', 'request'

    tick:    Tick
    request:  'FramedRequestMsg'

    def __init__(self, tick: Tick, request: 'FramedRequestMsg'):
        self.tick = tick
        self.request = request

    def __short_args__(self):
        yield self.request.short_str()

    def __line_args__(self):
        yield self.request.repr()


class OutgoingFrameEnter(FrameEnterEvent):
    __slots__ = 'tick', 'request'
    __icon__  = '··>'


class IncomingFrameEnter(FrameEnterEvent):
    __slots__ = 'tick', 'request'
    __icon__  = '<··'


################################################################################

class FrameExitEvent(FrameEvent):
    __slots__ = 'start', 'tick', 'request', 'response'

    start:    Tick
    request:  'FramedRequestMsg'

    tick:     Tick
    response: 'FramedResponseMsg'

    def __init__(self, start: Tick, request: 'FramedRequestMsg', tick: Tick, response: 'FramedResponseMsg'):
        self.start = start
        self.tick = tick
        self.request = request
        self.response = response

    def __short_args__(self):
        yield str(self.start)
        yield self.request.short_str()
        yield self.response.short_str()

    def __line_args__(self):
        # yield self.request.repr()
        yield self.response.repr()
        yield f'#{self.start}'


class OutgoingFrameExit(FrameExitEvent):
    __slots__ = 'start', 'tick', 'request', 'response'
    __icon__  = '-->'


class IncomingFrameExit(FrameExitEvent):
    __slots__ = 'start', 'tick', 'request', 'response'
    __icon__  = '<--'


################################################################################

class EventEvent(Event):
    __slots__ = 'tick', 'event',

    tick:   Tick
    event: 'EventMsg'

    def __init__(self, tick: Tick, event: 'EventMsg'):
        self.tick = tick
        self.event = event

    def __short_args__(self):
        yield self.event.short_str()

    def __line_args__(self):
        yield self.event.repr()


class OutgoingEvent(EventEvent):
    __slots__ = 'tick', 'event',
    __icon__  = '#->'


class IncomingEvent(EventEvent):
    __slots__ = 'tick', 'event',
    __icon__  = '<-#'


################################################################################

class ResourceEvent(Event):
    __slots__ = 'tick', 'resource',

    tick:    Tick
    resource: 'ResourceData'

    def __init__(self, tick: Tick, resource: 'ResourceData'):
        self.tick = tick
        self.resource = resource

    def __short_args__(self):
        yield self.resource.short_str()

    def __line_args__(self):
        yield self.resource.repr()


class OutgoingResource(ResourceEvent):
    __slots__ = 'tick', 'resource',
    __icon__  = '@->'


class IncomingResource(ResourceEvent):
    __slots__ = 'tick', 'resource',
    __icon__  = '<-@'


################################################################################

class MessageEvent(Event):
    __slots__ = 'tick', 'msg',

    tick:  Tick
    msg:  'RPCMsg'

    def __init__(self, tick: Tick, msg: 'RPCMsg'):
        self.tick = tick
        self.msg = msg

    def __short_args__(self):
        yield self.msg.short_str()

    def __line_args__(self):
        yield self.msg.repr()


class OutgoingMessage(Event):
    __slots__ = 'tick', 'msg',
    __icon__  = '*->'


class IncomingMessage(Event):
    __slots__ = 'tick', 'msg',
    __icon__  = '<-*'


################################################################################

FrameEnterEvent.INCOMING = IncomingFrameEnter
FrameEnterEvent.OUTGOING = OutgoingFrameEnter

FrameExitEvent.INCOMING = IncomingFrameExit
FrameExitEvent.OUTGOING = OutgoingFrameExit

ResourceEvent.INCOMING = IncomingResource
ResourceEvent.OUTGOING = OutgoingResource

MessageEvent.INCOMING = IncomingMessage
MessageEvent.OUTGOING = OutgoingMessage

EventEvent.INCOMING = IncomingEvent
EventEvent.OUTGOING = OutgoingEvent



from .base import AgenticaError, enrich_error
from .bugs import *
from .connection import *
from .generation import *
from .invocation import *

# TODO: put multiplex errors here too...

__all__ = [
    # base exception
    'AgenticaError',
    'enrich_error',
    # connection errors
    'ConnectionError',
    'WebSocketConnectionError',
    'WebSocketTimeoutError',
    'SDKUnsupportedError',
    'ClientServerOutOfSyncError',
    # server/generation errors
    'ServerError',
    'APIConnectionError',
    'APITimeoutError',
    'BadRequestError',
    'ConflictError',
    'ContentFilteringError',
    'DeadlineExceededError',
    'GenerationError',
    'InferenceError',
    'InternalServerError',
    'UsageError',
    'MaxTokensError',
    'MaxRoundsError',
    'NotFoundError',
    'OverloadedError',
    'PermissionDeniedError',
    'InsufficientCreditsError',
    'RateLimitError',
    'RequestTooLargeError',
    'ServiceUnavailableError',
    'UnauthorizedError',
    'UnprocessableEntityError',
    # invocation errors
    'InvocationError',
    'TooManyInvocationsError',
    'NotRunningError',
    # bugs
    'ThisIsABug',
]



# fmt: off

__all__ = [
    'PropClassSlot',
    'PropClassConst',
    'PropClassMutable',
    'WITH_PROP_CLASSES',
    'WITH_PROP_INSTANCES',
]


class PropClassSlot:
    """PropClassSlot docstring"""

    __slots__ = ['slot_prop']


class PropClassConst:
    """PropClassConst docstring"""

    @property
    def const_prop(self) -> int:
        return 0


class PropClassMutable:
    """PropClassMutable docstring"""

    _val: str

    @property
    def mut_prop(self) -> str:
        return self._val

    @mut_prop.setter
    def _(self, value: str) -> None:
        self._val = value

    @mut_prop.setter
    def _(self, value: str) -> None:
        del self._val


WITH_PROP_CLASSES: list[type] = [
    PropClassSlot,
    PropClassConst,
    PropClassMutable,
]

WITH_PROP_INSTANCES: list[object] = [
    PropClassSlot(),
    PropClassConst(),
    PropClassMutable(),
]



# fmt: off

import asyncio
from typing import ClassVar

from agentica_internal.testing import *
from agentica_internal.warpc.worlds.debug_world import *


class GenericClass[S, T]:
    pass


class UserClass:
    n: ClassVar[int] = 0
    a: int
    b: str
    _c: float

    X = 55

    def __init__(self, a: int, b: str, c: float):
        UserClass.n += 1
        self.a = a
        self.b = b
        self.c = c

    def user_method(self):
        return UserClass.n, self.a, self.b, self.c

    def ___forbidden_method___(self):
        ...


async def verify_generic_class():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair as B:
        a_cls = GenericClass
        b_cls = await B(a_cls)
        a_tp = a_cls.__type_params__
        b_tp = b_cls.__type_params__
        a_tp_types = tuple(type(t) for t in a_tp)
        b_tp_types = tuple(type(t) for t in b_tp)
        a_alias = a_cls[int, float]
        b_alias = b_cls[int, float]
        assert a_tp_types == b_tp_types, f"{a_tp_types} != {b_tp_types}"
        assert f_anno(a_tp) == f_anno(b_tp), f"{a_tp} != {b_tp}"
        assert f_anno(a_alias) == f_anno(b_alias), f"{a_alias} != {b_alias}"

    return pair


async def verify_user_class():

    pair = DebugWorld.connected_pair(logging=False, qualify_names=False)
    async with pair:
        A, B = pair.pipes

        a_s = UserClass(5, 'abc', 3.5)
        a_s.x = None
        b_s = await B(a_s)

        assert a_s.a == b_s.a
        assert a_s.b == b_s.b
        assert a_s.c == b_s.c
        assert a_s.x == b_s.x

        b_cls = await B(UserClass)
        b_obj = b_cls(5, 'abc', 3.5)
        assert is_virtual(b_cls)
        a_obj = await A(b_obj)
        assert is_real(a_obj)
        assert is_virtual(b_cls.__dict__.get('X'))  # a virtual property
        assert UserClass.X == 55
        assert b_cls.X == 55
        # these cannot actually work, since class-level descriptors do not intercept mutation
        # b_cls.X = 99
        # assert UserClass.X == 99
        assert is_virtual(b_cls.user_method)
        assert is_real(b_obj.user_method)
        assert b_obj.user_method is not a_obj.user_method
        a_result = a_obj.user_method()
        b_result = b_obj.user_method()
        assert a_result == b_result

        assert type(a_obj) is UserClass, f"{type(a_obj)} != UserClass"
        assert a_obj.a == 5
        assert a_obj.b == 'abc'
        assert a_obj.c == 3.5

    return pair


if __name__ == '__main__':
    asyncio.run(verify_generic_class())
    asyncio.run(verify_user_class())



# fmt: off

from asyncio import Queue
from pathlib import Path

from ...core.debug import enable_rich_tracebacks
from ...core.log import LoggingSpec, set_log_tags
from .__ import *

__all__ = ['ConnectedPair']


################################################################################

if TYPE_CHECKING:
    from .debug_world import DebugWorld

################################################################################


class Sending(Awaitable):
    __slots__ = 'src', 'dst', 'real', 'virt'

    def __init__(self, src, dst, real):
        self.src = src
        self.dst = dst
        self.real = real

    def __await__(self):
        async def send():
            if hasattr(self, 'virt'):
                return self.virt
            await self.src.channel_send_value(self.real)
            self.virt = virt = await self.dst.channel_recv_value()
            return virt

        return send().__await__()

    def __call__(self, *args):
        assert callable(self.real)

        async def call(orig_args: tuple):
            sent_fn = await self
            sent_args = []
            for orig in orig_args:
                sent = orig
                if isinstance(orig, Sending):
                    if orig.src is self.src:
                        sent = orig.real
                    else:
                        sent = await orig
                sent_args.append(sent)
            return sent_fn(*sent_args)

        return call(args)


################################################################################


class ConnectedPair:
    a: 'DebugWorld'
    b: 'DebugWorld'
    _logging: LoggingSpec
    _tmp_file: Path | None

    def __init__(
        self,
        a_name: str = 'a',
        b_name: str = 'b',
        logging: LoggingSpec = None,
        dump_msgs: bool = False,
        **kwargs,
    ) -> None:
        from .debug_world import DebugWorld

        enable_rich_tracebacks()
        P.NOW_FORMAT = ''
        self._reset_logging = set_log_tags(logging)
        self._logging = logging
        self.a = a = DebugWorld(a_name, **kwargs)
        self.b = b = DebugWorld(b_name, **kwargs)
        self.l = logging
        self._tmp_file = None
        a.other = b
        b.other = a
        if dump_msgs:
            a_path = Path('debug_world_a.msgs').absolute()
            b_path = Path('debug_world_b.msgs').absolute()
            print(f"writing debug world msgs to:\n{a_path}\n{b_path}")
            a.write_msgs_to(a_path)
            b.write_msgs_to(b_path)

    @property
    def tmp_file(self) -> Path:
        from tempfile import mktemp

        if file := self._tmp_file:  # type: ignore
            return file
        self._tmp_file = file = Path(mktemp())
        return file

    @property
    def worlds(self):
        return self.a, self.b

    @property
    def pipes(self):
        return self.A, self.B

    def B(self, v):
        return Sending(self.a, self.b, v)

    def A(self, v):
        return Sending(self.b, self.a, v)

    __call__ = B

    async def __aenter__(self):
        set_log_tags(self._logging)
        a_to_b = Queue()
        b_to_a = Queue()
        self.a.start_msg_loop(a_to_b.put, b_to_a.get)
        self.b.start_msg_loop(b_to_a.put, a_to_b.get)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.a.close()
        self.b.close()
        self._reset_logging()
        if file := self._tmp_file:
            file.unlink(missing_ok=True)

    def history_str(self) -> str:
        a_name = self.a.log_name
        b_name = self.b.log_name
        a_str = self.a.history_str()
        b_str = self.b.history_str()
        return f'{a_name}:\n{a_str}\n\n{b_name}:\n{b_str}'

    def print(self):
        print(self.history_str())

    def collect_events(self, events: list[Event]):
        self.a.collect_events(events)
        self.b.collect_events(events)

    @property
    def last_a_event(self) -> Event:
        return self.a.events[-1]

    @property
    def last_b_event(self) -> Event:
        return self.b.events[-1]



from types import FunctionType

from agentica_internal.core.mixin import bases, init_module

__all__ = 'foo', 'bar', 'get_registry', 'is_feature_a'


@init_module
def init_a(mod) -> None:
    print(f'{mod.__name__.split(".")[-1]}.__init_submodue___')
    mod.x *= 10


x = 3


def foo() -> str:
    return 'a.foo'


def bar() -> tuple[str, str]:
    return 'a.bar', foo()


def is_feature_a() -> bool:
    return True


def get_registry() -> dict[str, str]:
    """
    dynamically compute registry from current module's is_*/has_* functions.

    globals() returns the rewritten globals (child module's namespace).
    bases() figures out the current module from caller's globals.
    """
    registry = {}

    # inherit from base modules
    for base in bases():
        registry |= base.get_registry()

    # scan current module for is_*/has_* functions
    for name, value in globals().items():
        if isinstance(value, FunctionType):
            if name.startswith(('is_', 'has_')):
                registry[name] = name

    return registry



from typing import Any

from ...core.fmt import f_tag_str

__all__ = [
    'ObjectNamer',
]


class ObjectNamer:
    str_to_obj: dict[str, int]

    def __init__(self, path_safe: bool = True):
        self.str_to_obj = {}
        self.path_safe = path_safe

    def __call__(self, obj: Any) -> str:
        tag, name = f_tag_str(obj)
        name = name.replace('collections.abc.', 'A.').replace('typing.', 'T.')
        name = f'{tag}:{name}'
        prev = self.str_to_obj.get(name, not_present)
        if prev is not_present:
            self.str_to_obj[name] = prev
        else:
            assert prev is obj
        return name


not_present = object()



from agentica_internal.repl.repl import *


def verify_explicit_raise():
    repl = BaseRepl()
    error = repl.run_code("raise KeyError(1)").error
    assert error
    assert error.raised
    assert isinstance(error.exception, KeyError)


def verify_ignores_reraise():
    repl = BaseRepl()
    error = repl.run_code('''
    try:
        1/0
    except:
        raise
    ''').error
    assert error
    assert not error.raised
    assert isinstance(error.exception, ZeroDivisionError)


if __name__ == '__main__':
    verify_explicit_raise()
    verify_ignores_reraise()



# fmt: off

from enum import StrEnum

__all__ = [
    'Vars',
    'VarKeys',
    'Options',
    'Scope',
    'ReplException',
    'LOCALS',
    'GLOBALS',
    'BUILTINS',
    'HIDDEN',
    'USER',
    'REPL_VAR'
]

################################################################################

type Vars      = dict[str, object]
type VarKeys   = tuple[str, ...]
type Options   = dict[str, bool | int | str | float | None]

################################################################################

class Scope(StrEnum):
    USER     = 'user'
    LOCALS   = 'locals'
    GLOBALS  = 'globals'
    HIDDEN   = 'hidden'
    BUILTINS = 'builtins'

LOCALS   = Scope.LOCALS
GLOBALS  = Scope.GLOBALS
BUILTINS = Scope.BUILTINS
HIDDEN   = Scope.HIDDEN
USER     = Scope.USER

################################################################################

class REPL_VAR:
    ROLE             = '__role'
    TASK_DESCRIPTION = '__task_desc'
    SELF_FN          = '__self'
    RETURN_TYPE      = '__return_type'
    MAX_REPR_LEN     = '__max_repr_len'

################################################################################

class ReplException(BaseException):
    pass



import agentica_internal.cpython.ex.named as NAMED
from agentica_internal.cpython.type import CPythonObj, CPythonType

__all__ = ['NAMED', 'VALS', 'TYPES', 'DICT', 'cpython_type_example']


KEYS: list[str] = NAMED.__all__
VALS: list[CPythonObj] = [getattr(NAMED, k) for k in KEYS]
TYPES: list[CPythonType] = [type(v) for v in VALS]
DICT: dict[str, CPythonObj] = dict(zip(KEYS, VALS))

TO_EXAMPLE: dict[CPythonType, CPythonObj] = dict(zip(TYPES, VALS))


def cpython_type_example(cls: CPythonType) -> CPythonObj:
    return TO_EXAMPLE[cls]



# fmt: off

import ast
import builtins

from types import LambdaType

from .__ import *
from .base import Msg
from .term import TermPassByValMsg
from .term_atom import AtomMsg


__all__ = [
    'LambdaMsg',
    'SyntaxMsg',
]


################################################################################

class LambdaMsg(TermPassByValMsg):

    syntax:  'LambdaSyntax'
    globals: 'AttributesT'

    @staticmethod
    def encode_lambda(enc: EncoderP, func: FunctionType) -> 'LambdaMsg':
        return function_to_lambda_msg(enc, func)

    def decode(self, dec) -> FunctionType:
        return lambda_msg_to_function(dec, self)


################################################################################

type SyntaxMsgs = list[SyntaxMsg]

class SyntaxMsg(Msg): ...


################################################################################

class LambdaSyntax(SyntaxMsg, tag='lambda$'):

    args: list[str]
    body: 'SyntaxMsg'


################################################################################

class NameSyntax(SyntaxMsg):
    name: str

class LambdaArgSyntax(NameSyntax,   tag='arg$'): ...
class BuiltinNameSyntax(NameSyntax, tag='sys$'): ...
class GlobalNameSyntax(NameSyntax,  tag='var$'): ...

################################################################################

type BinOp = Literal[
    'Add', 'Sub', 'Mult', 'Div', 'FloorDiv', 'Mod', 'Pow', 'LShift', 'RShift',
    'BitOr', 'BitXor', 'BitAnd', 'MatMult'
]

class BinarySyntax(SyntaxMsg, tag='binary$'):
    left:  SyntaxMsg
    op:    BinOp
    right: SyntaxMsg


################################################################################

type BoolOp = Literal['A', 'USub', 'Not', 'Invert']

class BoolSyntax(SyntaxMsg):
    args: SyntaxMsgs

class AndSyntax(BoolSyntax, tag='and$'): ...
class OrSyntax(BoolSyntax, tag='or$'): ...


################################################################################

type CmpOp = Literal[
    'Eq', 'NotEq', 'Lt', 'LtE', 'Gt', 'GtE', 'Is', 'IsNot', 'In', 'NotIn'
]


class CompareSyntax(SyntaxMsg, tag='cmp$'):
    left:  SyntaxMsg
    op:    CmpOp
    right: SyntaxMsg


################################################################################

type UnaryOp = Literal['UAdd', 'USub', 'Not', 'Invert']

class UnarySyntax(SyntaxMsg, tag='unary$'):
    op:  UnaryOp
    arg: SyntaxMsg


################################################################################

class ConstantSyntax(SyntaxMsg, tag='const$'):
    const: AtomMsg


################################################################################

class CallSyntax(SyntaxMsg, tag='call$'):
    fn:   SyntaxMsg
    args: SyntaxMsgs


################################################################################

class TernaryIfSyntax(SyntaxMsg, tag='if$'):
    test:  SyntaxMsg
    tbody: SyntaxMsg
    fbody: SyntaxMsg


################################################################################

class AwaitSyntax(SyntaxMsg, tag='await$'):
    arg: SyntaxMsg


################################################################################

class GetAttrSyntax(SyntaxMsg, tag='getattr$'):
    obj:  SyntaxMsg
    attr: str


class GetItemSyntax(SyntaxMsg, tag='getitem$'):
    obj:  SyntaxMsg
    item: SyntaxMsg


################################################################################

class ContainerSyntax(SyntaxMsg):
    vals: SyntaxMsgs = []

class ListSyntax(ContainerSyntax,  tag='list$'): ...
class TupleSyntax(ContainerSyntax, tag='tuple$'): ...
class SetSyntax(ContainerSyntax,   tag='set$'): ...

class DictSyntax(ContainerSyntax, tag='dict$'):
    keys: SyntaxMsgs = []
    vals: SyntaxMsgs = []


################################################################################

class FmtSyntax(SyntaxMsg, tag='fmt$'):
    value: SyntaxMsg
    conversion: int

class FStringSyntax(SyntaxMsg, tag='fstr$'):
    elems: list[str | FmtSyntax]


################################################################################

# TODO: hook these up

class ComprehensionIteratorSyntax(SyntaxMsg, tag='comp_iter$'):
    target: SyntaxMsg
    iter:   SyntaxMsg
    tests:  SyntaxMsgs

class ComprehensionSyntax(SyntaxMsg):
    elem:  SyntaxMsg
    iters: list[ComprehensionIteratorSyntax]

class ListComprehensionSyntax(ComprehensionSyntax, tag='list_comp$'): ...
class SetComprehensionSyntax(ComprehensionSyntax, tag='set_comp$'): ...
class DictComprehensionSyntax(ComprehensionSyntax, tag='dict_comp$'): ...
class GeneratorSyntax(ComprehensionSyntax, tag='generator$'): ...


################################################################################

def function_to_lambda_msg(enc: EncoderP, func: FunctionType | LambdaType) -> LambdaMsg:
    from inspect import getsource
    try:
        source = getsource(func)
    except:
        source = None
    if not source:
        raise E.WarpLambdaEncodingError(f'could not obtain source for {func!r}')
    return source_to_lambda_msg(enc, source, func.__globals__)

def source_to_lambda_msg(enc: EncoderP, source: str, lambda_globals: AttributesT | None = None) -> LambdaMsg:
    from ast import parse
    source = source.strip()
    if not source.startswith('lambda ') and 'lambda ' not in source:
        raise E.WarpLambdaEncodingError(f'not a lambda function: {source!r}')
    try:
        tree = parse(source, '<lambda>', 'exec')
    except OSError:
        tree = None
    if tree is None:
        raise E.WarpLambdaEncodingError(f'could not recompile lambda:\n{source!r}\ngot {tree!r}')
    lambda_node = find_lambda_node(tree)
    if lambda_node is None:
        raise E.WarpLambdaEncodingError(f'could not find lambda:\n{source!r}')
    return encode_lambda_msg(enc, lambda_node, lambda_globals)

def find_lambda_node(node: ast.AST) -> ast.Lambda | None:
    # TODO: make this more precise, since there might be multiple lambdas in
    # the returned source. we can do better by 1) checking the argument names
    # to disambiguate 2) aborting if there is STILL possible ambiguity
    if isinstance(node, ast.Lambda):
        return node
    for node in ast.walk(node):
        if isinstance(node, ast.Lambda):
            return node
    return None

################################################################################

def lambda_msg_to_ast(dec: DecoderP, msg: LambdaMsg) -> ast.Lambda:
    lambda_node, lambda_globals = decode_lambda_msg(dec, msg)
    return lambda_node

def lambda_msg_to_source(dec: DecoderP, msg: LambdaMsg) -> str:
    lambda_node, lambda_globals = decode_lambda_msg(dec, msg)
    return ast.unparse(lambda_node)

def lambda_msg_to_function(dec: DecoderP, msg: LambdaMsg) -> FunctionType:
    lambda_node, lambda_globals = decode_lambda_msg(dec, msg)
    expr = ast.Expression(body=lambda_node)
    ast.fix_missing_locations(expr)
    try:
        code = compile(expr, '<lambda>', 'eval')
    except OSError:
        code = None
    if not isinstance(code, S.CodeT):
        f_node = ast.dump(lambda_node, indent=3)
        raise E.WarpLambdaDecodingError(f'could not compile lambda tree:\n{f_node}')
    try:
        func = eval(code, lambda_globals)
        if isinstance(func, FunctionType):
            return func
    except:
        pass

    raise E.WarpLambdaDecodingError(f'could not create a FunctionType from code object:\n{code}')

################################################################################

def encode_lambda_msg(encoder: EncoderP, root: ast.Lambda, known_globals: dict[str, Any]) -> LambdaMsg:

    lambda_globals: dict[str, Any] = {}
    lambda_args: set[str] = set()

    if known_globals is None:
        known_globals = {}
    get_global = known_globals.get

    def enc_syntax(node: ast.expr) -> SyntaxMsg:
        match node:
            case ast.Name(name, ast.Load()):
                return enc_name(name)

            case ast.Lambda():
                return enc_lambda(node)

            case ast.Constant(value):
                return ConstantSyntax(AtomMsg.encode_atom(value))

            case ast.UnaryOp(op, operand):
                return UnarySyntax(enc_unary_op(op), enc(operand))

            case ast.BinOp(left, op, right):
                return BinarySyntax(enc(left), enc_bin_op(op), enc(right))

            case ast.BoolOp(op, values):
                cls = AndSyntax if type(op) is ast.And else OrSyntax
                return cls(enc_list(values))

            case ast.Compare(left, ops, comparators):
                assert len(ops) == 1
                return CompareSyntax(enc(left), enc_cmp_op(ops[0]), enc(comparators[0]))

            case ast.Call(func, args, keywords):
                assert len(keywords) == 0
                return CallSyntax(enc(func), enc_list(args))

            case ast.IfExp(test, body, orelse):
                return TernaryIfSyntax(enc(test), enc(body), enc(orelse))

            case ast.Subscript(value, part, ast.Load()):
                return GetItemSyntax(enc(value), enc(part))

            case ast.Attribute(value, attr, ast.Load()):
                return GetAttrSyntax(enc(value), attr)

            case ast.Tuple(vals, ast.Load()):
                return TupleSyntax(enc_list(vals))
            case ast.List(vals, ast.Load()):
                return ListSyntax(enc_list(vals))
            case ast.Set(vals, ast.Load()):
                return SetSyntax(enc_list(vals))
            case ast.Dict(keys, vals):
                return DictSyntax(enc_list(keys), enc_list(vals))

            case ast.JoinedStr(values):
                return FStringSyntax([enc_fstr_node(v) for v in values])
            case ast.FormattedValue():
                return enc_fstr_node(node)

            case _:
                raise E.WarpLambdaEncodingError(node)

    def enc_list(nodes):
        return list(map(enc_syntax, nodes))

    def enc_fstr_node(elem) -> FmtSyntax | str:
        match elem:
            case ast.FormattedValue(fval, conversion, None):
                return FmtSyntax(enc(fval), conversion)
            case str():
                return elem
            case _:
                raise E.WarpLambdaEncodingError(elem)

    def enc_pos_args(args: ast.arguments) -> list[str]:
        match args:
            case ast.arguments(pos_args, reg_args, None, [], [], None, []):
                return [a.arg for a in pos_args + reg_args]
            case _:
                enc_node_error("unsupported ast.arguments", args)

    def enc_lambda(node: ast.Lambda) -> LambdaSyntax:
        args = enc_pos_args(node.args)
        lambda_args.update(args)
        body = enc_syntax(node.body)
        return LambdaSyntax(args, body)

    def enc_name(name: str) -> SyntaxMsg:
        if name in lambda_args:
            return LambdaArgSyntax(name)
        if name in lambda_globals:
            return lambda_globals[name]

        value = get_global(name, NO_GLOBAL)
        if value is not NO_GLOBAL:
            lambda_globals[name] = value
            return GlobalNameSyntax(name)

        if builtin_global := get_builtin(name):
            return builtin_global

        raise E.WarpLambdaEncodingError(f"cannot resolve name: {name!r}")

    enc = enc_syntax

    lambda_syntax_msg = enc_lambda(root)
    lambda_globals_msg = encoder.enc_record(lambda_globals)
    lambda_msg = LambdaMsg(lambda_syntax_msg, lambda_globals_msg)

    return lambda_msg


################################################################################

def decode_lambda_msg(decoder: DecoderP, root: LambdaMsg) -> tuple[ast.Lambda, AttributesT]:

    def dec_syntax(msg):
        match msg:

            case NameSyntax(name):
                return ast.Name(name, ast.Load())

            case LambdaSyntax():
                return dec_lambda(msg)

            case ConstantSyntax(const):
                return ast.Constant(const.decode_atom())

            case UnarySyntax(op, operand):
                return ast.UnaryOp(dec_unary_op(op), dec(operand))

            case BinarySyntax(left, op, right):
                return ast.BinOp(dec(left), dec_bin_op(op), dec(right))

            case AndSyntax(args):
                return ast.BoolOp(ast.And(), dec_list(args))

            case OrSyntax(args):
                return ast.BoolOp(ast.Or(), dec_list(args))

            case CompareSyntax(left, op, right):
                ops = [dec_cmp_op(op)]
                rights = [dec_syntax(right)]
                return ast.Compare(dec(left), ops, rights)

            case CallSyntax(func, args):
                return ast.Call(dec(func), dec_list(args), [])

            case TernaryIfSyntax(test, tbody, fbody):
                return ast.IfExp(dec(test), dec(tbody), dec(fbody))

            case GetItemSyntax(value, part):
                return ast.Subscript(dec(value), dec(part), ast.Load())

            case GetAttrSyntax(value, attr, ast.Load()):
                return ast.Attribute(dec(value), attr, ast.Load())

            case TupleSyntax(vals):
                return ast.Tuple(dec_list(vals), ast.Load())
            case ListSyntax(vals):
                return ast.List(dec_list(vals), ast.Load())
            case SetSyntax(vals):
                return ast.Set(dec_list(vals))
            case DictSyntax(keys, vals):
                return ast.Dict(dec_list(keys), dec_list(vals))

            case FStringSyntax(values):
                return ast.JoinedStr(list(map(dec_fstr_node, values)))
            case FmtSyntax():
                return dec_fstr_node(msg)

            case _:
                raise E.WarpLambdaDecodingError(msg)

    def dec_fstr_node(msg: FmtSyntax | str) -> ast.FormattedValue | str:
        match msg:
            case FmtSyntax(value, conversion):
                return ast.FormattedValue(dec_syntax(value), conversion)
            case str():
                return msg
            case _:
                raise E.WarpLambdaDecodingError(msg)

    def dec_pos_args(strs: list[str]) -> ast.arguments:
        args = [ast.arg(s) for s in strs]
        return ast.arguments([], args, None, [], [], None, [])

    def dec_lambda(msg: LambdaSyntax) -> ast.Lambda:
        args = dec_pos_args(msg.args)
        body = dec_syntax(msg.body)
        if not isinstance(body, ast.expr):
            raise E.WarpLambdaDecodingError()
        return ast.Lambda(args, body)

    def dec_list(msgs):
        return list(map(dec_syntax, msgs))

    dec = dec_syntax

    lambda_node = dec_lambda(root.syntax)
    lambda_globals = decoder.dec_record(root.globals)
    return lambda_node, lambda_globals


################################################################################

def enc_node_error(msg: str, node: ast.AST) -> NoReturn:
    f_node = ast.dump(node, indent=3)
    raise E.WarpLambdaEncodingError(f'{msg}\n{f_node}')


################################################################################

NO_GLOBAL = object()

################################################################################

builtin_globals: dict[str, BuiltinNameSyntax] = {
    name: BuiltinNameSyntax(name)
    for name in dir(builtins)
    if not name.startswith('__')
}

get_builtin = builtin_globals.get


################################################################################

def enc_dec_singletons[B](
    base: type[B],
    alias: T.TypeAliasType) -> tuple[Callable[[B], Any], Callable[[Any], B]]:
    value = alias.__value__
    names = value.__args__
    assert is_strtup(names)

    to_sng: dict[str, B] = {}
    for subcls in base.__subclasses__():
        name = subcls.__name__
        if name in names:
            to_sng[name] = subcls()

    def to_str(obj) -> str:
        name = type(obj).__name__
        assert name in names
        return name

    return to_str, to_sng.__getitem__

enc_bin_op, dec_bin_op     = enc_dec_singletons(ast.operator, BinOp)
enc_cmp_op, dec_cmp_op     = enc_dec_singletons(ast.cmpop, CmpOp)
enc_unary_op, dec_unary_op = enc_dec_singletons(ast.unaryop, UnaryOp)
